from .auth import (
    LoginView,
    LogoutView,
)
from .feature_flags import FeatureFlagsView

from .controller import ControllerListView

from .settings import SettingsView
from .sync import ContainerSyncRemoteView, ContainerSyncRegistryView

from .index_execution_environments import IndexRegistryEEView

__all__ = (
    # auth
    "LoginView",
    "LogoutView",

    # feature_flags
    "FeatureFlagsView",

    # controller
    "ControllerListView",

    # settings
    "SettingsView",

    # sync
    "ContainerSyncRemoteView",
    "ContainerSyncRegistryView",

    # index_execution_environments
    "IndexRegistryEEView",
)



from pulpcore.plugin import PulpPluginAppConfig


class PulpGalaxyPluginAppConfig(PulpPluginAppConfig):
    """Entry point for the galaxy plugin."""

    name = "galaxy_ng.app"
    label = "galaxy"
    version = "4.4.5"

    def ready(self):
        super().ready()
        from .signals import handlers  # noqa



import json
import logging
from urllib.parse import quote, urlencode

from django.core import exceptions
from django.utils.translation import gettext_lazy as _
from django.http.request import HttpRequest

from pulpcore.plugin.tasking import dispatch

from pulp_container.app import models as container_models

from galaxy_ng.app.api.ui import serializers
from galaxy_ng.app import models


log = logging.getLogger(__name__)


CATALOG_API = "https://catalog.redhat.com/api/containers/v1/repositories"


class CouldNotCreateContainerError(Exception):
    def __init__(self, remote_name, error=""):
        self.message = _(
            "Failed to create container {remote_name}. {error}".format(
                remote_name=remote_name, error=error)
        )

        super().__init__(self.message)


def _get_request(request_data):
    request = HttpRequest()

    for k, v in request_data.items():
        setattr(request, k, v)

    return request


def _parse_catalog_repositories(response_data):
    containers = []

    for container in response_data['data']:
        containers.append({
            "name": container['repository'],
            "upstream_name": container['repository'],
            "description": container['display_data']['short_description'],
            "readme": container['display_data']['long_description_markdown'],
        })

    return containers


def _update_distro_readme_and_description(distro, container_data):
    readme = models.ContainerDistroReadme.objects.get_or_create(container=distro)[0]
    readme.text = container_data['readme']
    readme.save()

    distro.description = container_data['description']
    distro.save()


def create_or_update_remote_container(container_data, registry_pk, request_data):
    # check if a distro matching the base name exists
    remote_repo_type = container_models.ContainerRepository.get_pulp_type()

    try:
        distro = container_models.ContainerDistribution.objects.get(
            base_path=container_data['name'])

        repo = distro.repository

        # If a distro matching the container name exists, and it's a remote repository and
        # it has a container remote attached to it that's associated with the selected
        # registry, update the container's readme and description.
        if repo.pulp_type == remote_repo_type and repo.remote is not None:
            try:
                remote_registry = models.ContainerRegistryRepos.objects.get(
                    repository_remote=repo.remote).registry
                if remote_registry and remote_registry.pk == registry_pk:
                    _update_distro_readme_and_description(distro, container_data)
                    return

            except exceptions.ObjectDoesNotExist:
                raise CouldNotCreateContainerError(
                    container_data['name'],
                    error=_(
                        "A remote container with this name already exists, "
                        "but is not associated with any registry.")
                )

        else:
            raise CouldNotCreateContainerError(
                container_data['name'],
                error=_("A local container with this name already exists.")
            )

    except exceptions.ObjectDoesNotExist:
        # If no distributions match the selected container, create one.
        request = _get_request(request_data)
        serializer = serializers.ContainerRemoteSerializer(
            data={
                "name": container_data['name'],
                "upstream_name": container_data['name'],
                "registry": str(registry_pk)
            }, context={"request": request}
        )

        try:
            serializer.is_valid(raise_exception=True)
        except serializers.ValidationError as e:
            CouldNotCreateContainerError(
                container_data['name'],
                error=str(e)
            )
        serializer.create(serializer.validated_data)

        distro = container_models.ContainerDistribution.objects.get(
            base_path=container_data['name'])

        _update_distro_readme_and_description(distro, container_data)


def index_execution_environments_from_redhat_registry(registry_pk, request_data):
    registry = models.ContainerRegistryRemote.objects.get(pk=registry_pk)
    remotes = []

    query = {
        "filter": "build_categories=in=('Automation execution environment')",
        "page": 0,
        "sort_by": "creation_date[asc]"
    }

    # Iterate through the pages on the API response and add the relevant information
    # about each container repository.
    while True:
        url = CATALOG_API + "?" + urlencode(query, quote_via=quote)
        downloader = registry.get_downloader(url=url)
        download_result = downloader.fetch()
        with open(download_result.path) as fd:
            data = json.load(fd)
            remotes = remotes + _parse_catalog_repositories(data)
            if len(data['data']) == data['page_size']:
                query['page'] += 1
            else:
                break

    for remote in remotes:
        # create a subtask for each remote, so that if one fails, we can throw a usable error
        # message for the user to look at and prevent the rest of the repositories from failing.
        dispatch(
            create_or_update_remote_container,
            kwargs={
                "container_data": remote,
                "registry_pk": registry.pk,
                "request_data": request_data
            },
            exclusive_resources=["/api/v3/distributions/"]
        )



# Generated by Django 2.2.16 on 2020-10-27 19:15

from django.db import migrations, models
import django.db.models.deletion

DELETE_ORPHAN_GALAXY_COLLECTION_IMPORTS = """
DELETE
FROM galaxy_collectionimport
WHERE task_id in
    (SELECT task_id
     FROM galaxy_collectionimport
     WHERE task_id not in
         (SELECT task_id
          FROM ansible_collectionimport));
"""

class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0010_add_staging_rejected_repos'),
    ]

    operations = [
        migrations.RunSQL(sql=DELETE_ORPHAN_GALAXY_COLLECTION_IMPORTS,
                          reverse_sql=migrations.RunSQL.noop),
        migrations.AlterField(
            model_name='collectionimport',
            name='task_id',
            field=models.OneToOneField(db_column='task_id', on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='galaxy_import', serialize=False, to='ansible.CollectionImport'),
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-03 19:06

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('container', '0020_update_push_repo_perms'),
        ('galaxy', '0022_enforce_retain_repo_versions'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='containersynctask',
            name='repository',
        ),
        migrations.RemoveField(
            model_name='containersynctask',
            name='task',
        ),
        migrations.AlterField(
            model_name='containerregistryrepos',
            name='repository_remote',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='registry', serialize=False, to='container.containerremote'),
        ),
        migrations.DeleteModel(
            name='CollectionSyncTask',
        ),
        migrations.DeleteModel(
            name='ContainerSyncTask',
        ),
    ]



"""
Signal handlers for the Galaxy application.
Those signals are loaded by
galaxy_ng.app.__init__:PulpGalaxyPluginAppConfig.ready() method.
"""
from django.apps import apps
from django.dispatch import receiver
from django.db.models.signals import post_save, post_migrate
from pulp_ansible.app.models import AnsibleDistribution, AnsibleRepository, Collection
from galaxy_ng.app.access_control.statements import PULP_CONTAINER_VIEWSETS
from galaxy_ng.app.models import ContentRedirectContentGuard, Namespace


@receiver(post_save, sender=AnsibleRepository)
def ensure_retain_repo_versions_on_repository(sender, instance, created, **kwargs):
    """Ensure repository has retain_repo_versions set when created.
    retain_repo_versions defaults to 1 when not set.
    """

    if created and instance.retain_repo_versions is None:
        instance.retain_repo_versions = 1
        instance.save()


@receiver(post_save, sender=AnsibleDistribution)
def ensure_content_guard_exists_on_distribution(sender, instance, created, **kwargs):
    """Ensure distribution have a content guard when created."""

    content_guard = ContentRedirectContentGuard.objects.first()

    if created and instance.content_guard is None:
        instance.content_guard = content_guard
        instance.save()


@receiver(post_save, sender=Collection)
def create_namespace_if_not_present(sender, instance, created, **kwargs):
    """Ensure Namespace object exists when Collection object saved.
    django signal for pulp_ansible Collection model, so that whenever a
    Collection object is created or saved, it will create a Namespace object
    if the Namespace does not already exist.
    Supports use case: In pulp_ansible sync, when a new collection is sync'd
    a new Collection object is created, but the Namespace object is defined
    in galaxy_ng and therefore not created. This signal ensures the
    Namespace is created.
    """

    Namespace.objects.get_or_create(name=instance.namespace)


def set_pulp_container_access_policies(sender, **kwargs):
    apps = kwargs.get("apps")
    if apps is None:
        from django.apps import apps
    AccessPolicy = apps.get_model("core", "AccessPolicy")

    print("Overriding pulp_container access policy")
    for view in PULP_CONTAINER_VIEWSETS:
        policy, created = AccessPolicy.objects.update_or_create(
            viewset_name=view, defaults={**PULP_CONTAINER_VIEWSETS[view], "customized": True})


post_migrate.connect(
    set_pulp_container_access_policies,
    sender=apps.get_app_config("galaxy"),
    dispatch_uid="override_pulp_container_access_policies"
)



from rest_framework import serializers
from pulp_ansible.app import models as pulp_models


class RepositorySerializer(serializers.ModelSerializer):
    content_count = serializers.SerializerMethodField()

    class Meta:
        model = pulp_models.AnsibleRepository
        fields = (
            'name',
            'description',
            'pulp_id',
            'pulp_last_updated',
            'content_count'
        )

    def get_content_count(self, repo) -> int:
        return repo.latest_version().content.count()


class DistributionSerializer(serializers.ModelSerializer):
    repository = RepositorySerializer()

    class Meta:
        model = pulp_models.AnsibleDistribution
        fields = (
            'pulp_id',
            'name',
            'base_path',
            'repository'
        )



from django.urls import include, path
from rest_framework import routers

from . import views, viewsets

router = routers.SimpleRouter()
router.register('namespaces', viewsets.NamespaceViewSet, basename='namespaces')

namespace_urls = [
    path("", include(router.urls)),
]

auth_urls = [
    path("auth/token/", views.TokenView.as_view(), name="auth-token"),
]

# these are included in the base router so that they only appear under `content/<distro>/v3`
sync_urls = [
    path(
        "sync/",
        views.SyncRemoteView.as_view(),
        name='sync'
    ),
    path(
        "sync/config/",
        viewsets.SyncConfigViewSet.as_view({"get": "retrieve", "put": "update"}),
        name="sync-config",
    ),
]

urlpatterns = [
    # The following endpoints are related to issue https://issues.redhat.com/browse/AAH-224
    # For now endpoints are temporary deactivated
    #
    # path("", viewsets.RepoMetadataViewSet.as_view({"get": "retrieve"}), name="repo-metadata"),
    # path(
    #     "collections/all/",
    #     viewsets.UnpaginatedCollectionViewSet.as_view({"get": "list"}),
    #     name="all-collections-list",
    # ),
    # path(
    #     "collection_versions/all/",
    #     viewsets.UnpaginatedCollectionVersionViewSet.as_view({"get": "list"}),
    #     name="all-collection-versions-list",
    # ),
    path(
        "collections/", viewsets.CollectionViewSet.as_view({"get": "list"}), name="collections-list"
    ),
    path(
        "collections/<str:namespace>/<str:name>/",
        viewsets.CollectionViewSet.as_view(
            {"get": "retrieve", "patch": "update", "delete": "destroy"}
        ),
        name="collections-detail",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/",
        viewsets.CollectionVersionViewSet.as_view({"get": "list"}),
        name="collection-versions-list",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/",
        viewsets.CollectionVersionViewSet.as_view({"get": "retrieve", "delete": "destroy"}),
        name="collection-versions-detail",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/docs-blob/",
        viewsets.CollectionVersionDocsViewSet.as_view({"get": "retrieve"}),
        name="collection-versions-detail-docs",
    ),
    path(
        "imports/collections/<str:pk>/",
        viewsets.CollectionImportViewSet.as_view({"get": "retrieve"}),
        name="collection-import",
    ),
    path(
        "artifacts/collections/",
        viewsets.CollectionUploadViewSet.as_view({"post": "create"}),
        name="collection-artifact-upload",
    ),
    path(
        "artifacts/collections/<str:path>/<str:filename>",
        viewsets.CollectionArtifactDownloadView.as_view(),
        name="collection-artifact-download",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/move/"
        "<str:source_path>/<str:dest_path>/",
        viewsets.CollectionVersionMoveViewSet.as_view({"post": "move_content"}),
        name="collection-version-move",
    ),
    path("tasks/", viewsets.TaskViewSet.as_view({"get": "list"}), name="tasks-list"),
    path("tasks/<str:pk>/", viewsets.TaskViewSet.as_view({"get": "retrieve"}), name="tasks-detail"),
    path("excludes/", views.ExcludesView.as_view(), name="excludes-file"),
]



import logging
from unittest import mock

from django.conf import settings
from guardian import shortcuts

from pulp_ansible.app import models as pulp_ansible_models

from galaxy_ng.app import models as galaxy_models
from galaxy_ng.app.models import auth as auth_models

from . import base

log = logging.getLogger(__name__)

ACCOUNT_SCOPE = "rh-identity-account"

SYNCLIST_PERMS = [
    "add_synclist",
    "view_synclist",
    "delete_synclist",
    "change_synclist",
]


log.info("settings.FIXTURE_DIRS(module scope): %s", settings.FIXTURE_DIRS)


class BaseSyncListViewSet(base.BaseTestCase):
    url_name = "galaxy:api:v3:ui:synclists-list"
    default_owner_permissions = SYNCLIST_PERMS

    def setUp(self):
        super().setUp()

        self.repo = self._create_repository("test_post_repo")
        self.repo.save()

        repo_name = settings.GALAXY_API_DEFAULT_DISTRIBUTION_BASE_PATH
        self.default_repo, _ = pulp_ansible_models.AnsibleRepository.objects.get_or_create(
            name=repo_name
        )

        self.default_dist, _ = pulp_ansible_models.AnsibleDistribution.objects.get_or_create(
            name=repo_name, base_path=repo_name, repository=self.default_repo
        )

        from galaxy_ng.app.access_control.statements import INSIGHTS_STATEMENTS  # noqa
        patcher = mock.patch(
            "galaxy_ng.app.access_control." "access_policy.AccessPolicyBase._get_statements",
            return_value=INSIGHTS_STATEMENTS,
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    @staticmethod
    def _create_group_with_synclist_perms(scope, name, users=None):
        group, _ = auth_models.Group.objects.get_or_create_identity(scope, name)
        if isinstance(users, auth_models.User):
            users = [users]
        group.user_set.add(*users)

        for perm in SYNCLIST_PERMS:
            shortcuts.assign_perm(f"galaxy.{perm}", group)
        return group

    def _create_repository(self, name):
        repo, _ = pulp_ansible_models.AnsibleRepository.objects.get_or_create(name="test_repo1")
        return repo

    def _create_synclist(
        self,
        name,
        repository,
        collections=None,
        namespaces=None,
        policy=None,
        users=None,
        groups=None,
        upstream_repository=None,
    ):
        upstream_repository = upstream_repository or self.default_repo
        if isinstance(groups, auth_models.Group):
            groups = [groups]
        else:
            groups = groups or []

        groups_to_add = {}
        for group in groups:
            groups_to_add[group] = self.default_owner_permissions

        synclist, _ = galaxy_models.SyncList.objects.get_or_create(
            name=name, repository=repository, upstream_repository=upstream_repository,
        )

        synclist.groups = groups_to_add
        synclist.save()

        return synclist



# Generated by Django 2.2.15 on 2020-08-11 19:44

from django.db import migrations
from django.contrib.auth.management import create_permissions


# note: using assign_perm from guardian.shortcuts doesn't work in migrations
# https://github.com/django-guardian/django-guardian/issues/281#issuecomment-156264129
# The following function is adapted from https://gist.github.com/xuhcc/67871719116bdc0fee6c
def assign_perm(apps, perm, owner, obj=None):
    Permission = apps.get_model('auth', 'Permission')
    UserObjectPermission = apps.get_model('guardian', 'UserObjectPermission')
    GroupObjectPermission = apps.get_model('guardian', 'GroupObjectPermission')
    app_label, codename = perm.split('.', 1)

    perm = Permission.objects.get(
        content_type__app_label=app_label,
        codename=codename)
    kwargs = {
        'permission': perm,
        'content_type': perm.content_type,
        'object_pk': obj.pk,
    }
    kwargs['group'] = owner
    return GroupObjectPermission.objects.get_or_create(**kwargs)



def convert_namespace_groups_to_permissions(apps, schema_editor):
    NamespaceModel = apps.get_model('galaxy', 'Namespace')

    for namespace in NamespaceModel.objects.all():
        for group in namespace.groups.all():
            assign_perm(apps, 'galaxy.change_namespace', group, namespace)
            assign_perm(apps, 'galaxy.upload_to_namespace', group, namespace)


def convert_synclist_groups_to_permissions(apps, schema_editor):
    SyncListModel = apps.get_model('galaxy', 'SyncList')

    for synclist in SyncListModel.objects.all():
        for group in synclist.groups.all():
            assign_perm(apps, 'galaxy.change_synclist', group, synclist)
            assign_perm(apps, 'galaxy.view_synclist', group, synclist)


# Ensures that the new permissions are added to the database so that they can
# be assigned during migrations
# based off of https://stackoverflow.com/a/40092780
def initialize_permissions(apps, schema_editor):
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, apps=apps, verbosity=0)
        app_config.models_module = None


class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0003_inbound_repo_per_namespace'),
        ('guardian', '0002_generic_permissions_index'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='namespace',
            options={'permissions': (('upload_to_namespace', 'Can upload collections to namespace'),)},
        ),
        migrations.RunPython(
            code=initialize_permissions,
        ),
        migrations.RunPython(
            code=convert_namespace_groups_to_permissions,
        ),
        migrations.RemoveField(
            model_name='namespace',
            name='groups',
        ),
        migrations.RunPython(
            code=convert_synclist_groups_to_permissions,
        ),
        migrations.RemoveField(
            model_name='synclist',
            name='groups',
        ),
        migrations.RemoveField(
            model_name='synclist',
            name='users',
        ),
    ]



# Generated by Django 3.2.6 on 2021-08-24 17:01

from django.db import migrations


def enforce_retain_repo_versions(apps, schema_editor):
    """
    Enforce that all repositories have retain_repo_versions=1
    """

    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')

    AnsibleRepository.objects.filter(
        retain_repo_versions=None
    ).update(
        retain_repo_versions=1
    )


class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0021_alter_user_first_name'),
    ]

    operations = [
        migrations.RunPython(
            code=enforce_retain_repo_versions,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.conf import settings

from .base import BaseTestCase, get_current_ui_url


class TestUiFeatureFlagsView(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.original_setting = settings.GALAXY_FEATURE_FLAGS
        settings.CONNECTED_ANSIBLE_CONTROLLERS = [
            "https://my-controller1.example.com/",
            "https://controllerino.contotrolisimo.com/",
            "https://boring-url.example.com/",
        ]

    def tearDown(self):
        super().tearDown()
        settings.CONNECTED_ANSIBLE_CONTROLLERS = self.original_setting

    def test_feature_flags_url(self):
        self.feature_flags_url = get_current_ui_url('controllers')

        controller_list = self.client.get(self.feature_flags_url).data['data']
        self.assertEqual(len(controller_list), len(settings.CONNECTED_ANSIBLE_CONTROLLERS))
        for controller in controller_list:
            self.assertTrue(controller['host'] in settings.CONNECTED_ANSIBLE_CONTROLLERS)

        controller_list = self.client.get(self.feature_flags_url + "?limit=1").data['data']
        self.assertEqual(len(controller_list), 1)

        controller_list = self.client.get(
            self.feature_flags_url + "?host__icontains=EXAMPLE.COM").data['data']
        self.assertEqual(len(controller_list), 2)
        for controller in controller_list:
            self.assertTrue('example.com' in controller['host'])

        controller_list = self.client.get(
            self.feature_flags_url + "?host=https://boring-url.example.com/").data['data']
        self.assertEqual(len(controller_list), 1)
        self.assertEqual(controller_list[0]["host"], "https://boring-url.example.com/")



from django.conf import settings
from rest_framework.permissions import IsAuthenticatedOrReadOnly
from rest_framework.response import Response

from galaxy_ng.app.api import base as api_base


class FeatureFlagsView(api_base.APIView):
    permission_classes = [IsAuthenticatedOrReadOnly]

    def get(self, request, *args, **kwargs):
        return Response(settings.GALAXY_FEATURE_FLAGS)



import logging
from rest_framework import status
from django.test import override_settings

from galaxy_ng.app.models import auth as auth_models

from galaxy_ng.app.constants import DeploymentMode

from .base import BaseTestCase, get_current_ui_url

log = logging.getLogger(__name__)


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestContainerRegistryRemoteViewset(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.admin_user = auth_models.User.objects.create(username='admin',
                                                          is_superuser=True)
        self.pe_group = self._create_partner_engineer_group()
        self.admin_user.groups.add(self.pe_group)
        self.admin_user.save()

    def test_get_registry_list_empty(self):
        url = get_current_ui_url('execution-environments-registry-list')
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(url)
        self.assertEqual(response.data['data'], [])

    def create_registry_remote(self, name):
        url = get_current_ui_url('execution-environments-registry-list')
        new_registry = {
            'name': name,
            'url' : 'http://example.com',
        }
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.post(url, new_registry, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        return response.data['pk']

    def test_get_detail_with_data(self):
        name = 'my_registry'
        pk = self.create_registry_remote(name)
        url = get_current_ui_url('execution-environments-registry-detail', kwargs={'pk': pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], name)

    def test_get_detail_fail(self):
        pk = 'this key does not exist'
        url = get_current_ui_url('execution-environments-registry-detail', kwargs={'pk': pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_get_list_with_data(self):
        test_name_1 = 'test_registry_one'
        test_name_2 = 'test_registry_two'
        self.create_registry_remote(test_name_1)
        self.create_registry_remote(test_name_2)
        url = get_current_ui_url('execution-environments-registry-list')
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(url)
        self.assertEqual(response.data['data'][0]['name'], test_name_1)
        self.assertEqual(response.data['data'][1]['name'], test_name_2)

    def test_update_detail(self):
        initial_name = 'intial_registry_name'
        updated_name = 'updated_registry_name'
        pk = self.create_registry_remote(initial_name)
        url = get_current_ui_url('execution-environments-registry-detail', kwargs={'pk': pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], initial_name)
        # logic for updating the registry
        self.client.put(url, {'name' : updated_name, 'url': 'https://example.com'}, format='json')
        response = self.client.get(url)
        self.assertEqual(response.data['name'], updated_name)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_detail_fail(self):
        initial_name = 'intial_registry_name'
        updated_name = 'updated_registry_name'
        pk = self.create_registry_remote(initial_name)
        url = get_current_ui_url('execution-environments-registry-detail', kwargs={'pk': pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], initial_name)
        # URL is required in PUT, so this should fail
        response = self.client.put(url, {'name' : updated_name}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_delete_registry_remote(self):
        name = 'registry_to_delete'
        pk = self.create_registry_remote(name)
        url = get_current_ui_url('execution-environments-registry-detail', kwargs={'pk': pk})
        # Get it first
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], name)
        # Delete it
        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        # Get it again
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



import enum


class DeploymentMode(enum.Enum):
    STANDALONE = 'standalone'
    INSIGHTS = 'insights'


CURRENT_UI_API_VERSION = 'v1'
ALL_UI_API_VERSION = {'v1': 'v1/'}

COMMUNITY_DOMAINS = (
    'galaxy.ansible.com',
    'galaxy-dev.ansible.com',
    'galaxy-qa.ansible.com',
)

INBOUND_REPO_NAME_FORMAT = "inbound-{namespace_name}"



import logging

from django.conf import settings
from django.db import transaction
from django.db.utils import IntegrityError
from django.utils.translation import gettext_lazy as _

from rest_framework.exceptions import ValidationError
from rest_framework import serializers

from pulp_ansible.app.models import Collection
from pulp_ansible.app.models import AnsibleRepository, AnsibleDistribution

from galaxy_ng.app import models
from galaxy_ng.app.access_control.fields import GroupPermissionField

log = logging.getLogger(__name__)

default_repo_name = settings.GALAXY_API_DEFAULT_DISTRIBUTION_BASE_PATH


class SyncListCollectionSummarySerializer(serializers.Serializer):
    namespace = serializers.CharField(max_length=64)
    name = serializers.CharField(max_length=64)


class SyncListSerializer(serializers.ModelSerializer):
    namespaces = serializers.SlugRelatedField(
        many=True, slug_field="name", queryset=models.Namespace.objects.all()
    )

    collections = SyncListCollectionSummarySerializer(many=True)

    groups = GroupPermissionField()

    def _get_repository(self, repository_id):
        try:
            repository = AnsibleRepository.objects.get(pulp_id=repository_id)
            return repository
        except AnsibleRepository.DoesNotExist:
            errmsg = _('Repository "{pulp_id}" not found while creating synclist')
            raise ValidationError(errmsg.format(pulp_id=repository_id))

    def to_internal_value(self, data):
        repository_data = data.get("repository", None)
        if repository_data:
            data["repository"] = self._get_repository(repository_data)

        upstream_repository_data = data.get("upstream_repository", None)
        if upstream_repository_data:
            data["upstream_repository"] = self._get_repository(upstream_repository_data)
        else:
            # If not specified, use the default upstream repo
            data["upstream_repository"] = AnsibleRepository.objects.get(name=default_repo_name)
        return super().to_internal_value(data)

    @transaction.atomic
    def _create_repository(self, name):
        repository, created = AnsibleRepository.objects.get_or_create(name=name)
        if created:
            repository.save()
        return repository

    def _update_distribution(self, name_slug, repository):
        # Now the distro
        try:
            distribution, create = AnsibleDistribution.objects.get_or_create(
                name=name_slug, base_path=name_slug,
            )
        except Exception as exc:
            log.exception(exc)
            raise

        # Now update the Distribution to point to the new Repository
        distribution.repository = repository

        distribution.save()
        return distribution

    @transaction.atomic
    def create(self, validated_data):
        collections_data = validated_data.pop("collections")
        namespaces_data = validated_data.pop("namespaces")
        repository = validated_data.pop("repository", None)
        name = validated_data.get("name")

        if not repository:
            repository = self._create_repository(name)

        try:
            instance = models.SyncList.objects.create(
                repository=repository,
                **validated_data
            )
        except IntegrityError as exc:
            raise ValidationError(_("Synclist already exists: %s") % exc)

        collections = []
        for collection_data in collections_data:
            try:
                collections.append(Collection.objects.get(**collection_data))
            except Collection.DoesNotExist:
                errmsg = (
                    _('Collection "{namespace}.{name}" not found '
                      'while creating synclist {synclist}')
                )
                raise ValidationError(
                    errmsg.format(
                        namespace=collection_data["namespace"],
                        name=collection_data["name"],
                        synclist=instance.name,
                    )
                )
        instance.collections.clear()
        instance.collections.set(collections)

        instance.namespaces.add(*namespaces_data)

        return instance

    @transaction.atomic
    def update(self, instance, validated_data):
        collections_data = validated_data.get("collections", [])

        groups_data = validated_data.get("groups")
        if groups_data:
            instance.groups = groups_data

        namespaces_data = validated_data.get("namespaces")
        if namespaces_data is not None:
            instance.namespaces.set(namespaces_data)

        instance.policy = validated_data.get("policy", instance.policy)

        instance.name = validated_data.get("name", instance.name)

        new_collections = []
        for collection_data in collections_data:
            try:
                new_collections.append(Collection.objects.get(**collection_data))
            except Collection.DoesNotExist:
                errmsg = (
                    _('Collection "{namespace}.{name}" not found '
                      'while updating synclist {synclist}')
                )
                raise ValidationError(
                    errmsg.format(
                        namespace=collection_data["namespace"],
                        name=collection_data["name"],
                        synclist=instance.name,
                    )
                )
        instance.collections.set(new_collections)

        # if collection goes from empty to having something in it,
        # then we also need to point the associated distro to the synclist repo
        if instance.collections.all() and instance.repository:
            self._update_distribution(instance.name, instance.repository)
        else:
            # synclist is no longer including/excluding anything, so point back
            # at the default upstream repo
            self._update_distribution(instance.name, instance.upstream_repository)

        instance.save()

        return instance

    class Meta:
        model = models.SyncList
        fields = [
            "id",
            "name",
            "policy",
            "upstream_repository",
            "repository",
            "collections",
            "namespaces",
            "groups",
        ]

        read_only_fields = ("repository", )



import io
import logging

from rest_framework.parsers import MultiPartParser

log = logging.getLogger(__name__)


class AnsibleGalaxy29MultiPartParser(MultiPartParser):
    def parse(self, stream, media_type=None, parser_context=None):
        # Add in a crlf newline if the body is missing it between the Content-Disposition line
        # and the value

        b_data = stream.read()
        body_parts = b_data.partition(b'Content-Disposition: form-data; name="sha256"\r\n')

        new_stream = io.BytesIO()
        new_stream.write(body_parts[0])
        new_stream.write(body_parts[1])

        if not body_parts[2].startswith(b'\r\n'):
            log.warning('Malformed multipart body user-agent: %s',
                        parser_context['request'].META.get('HTTP_USER_AGENT'))

            # add the crlf to the new stream so the base parser does the right thing
            new_stream.write(b'\r\n')

        new_stream.write(body_parts[2])

        new_stream.seek(0)

        return super(AnsibleGalaxy29MultiPartParser, self).parse(new_stream,
                                                                 media_type=media_type,
                                                                 parser_context=parser_context)



import logging
import itertools

from django.conf import settings
from django.db.models import F, Q
from django.utils.translation import gettext_lazy as _

from pulpcore.plugin.models import (
    CreatedResource,
    GroupProgressReport,
    ProgressReport,
    Task,
    TaskGroup,
)
from pulpcore.plugin.tasking import add_and_remove, dispatch
from pulp_ansible.app.models import (
    AnsibleRepository,
    CollectionVersion,
)

from galaxy_ng.app import models

log = logging.getLogger(__name__)


def curate_all_synclist_repository(upstream_repository_name, **kwargs):
    """When upstream_repository has changed, update all synclists repos associated with it.

    The synclist repos will be updated to upstream_repository

    This will create a lot of curate_synclist_repository tasks.
    It will create a TaskGroup containing those tasks.

    If neccasary, it may create many TaskGroups.

    It may need to schedule a series of TaskGroups, potentially
    in order of priority.

    This task need to be cancelable."""

    upstream_repository = AnsibleRepository.objects.get(name=upstream_repository_name)
    synclist_qs = models.SyncList.objects.filter(upstream_repository=upstream_repository)

    task_group = TaskGroup.objects.create(
        description=_("Curating all synclists repos that curate from %s") % upstream_repository_name
    )
    task_group.save()

    CreatedResource.objects.create(content_object=task_group)

    current_task = Task.current()
    current_task.task_group = task_group
    current_task.save()

    GroupProgressReport(
        message=_("Synclists curating upstream repo"),
        code="synclist.curate",
        total=synclist_qs.count(),
        task_group=task_group,
    ).save()

    with ProgressReport(
        message=_("Synclists curating upstream repo task"),
        code="synclist.curate.log",
        total=synclist_qs.count(),
    ) as task_progress_report:

        synclist_iter = synclist_qs.iterator()
        while True:
            batch = list(itertools.islice(synclist_iter, settings.SYNCLIST_BATCH_SIZE))
            if not batch:
                break

            # TODO: filter down to just synclists that have a synclist repo
            # locks need to be Model or str not int
            synclist_ids = [synclist.id for synclist in batch]
            locks = [synclist.repository for synclist in batch]

            dispatch(
                curate_synclist_repository_batch,
                locks,
                args=(synclist_ids,),
                task_group=task_group,
            )
            task_progress_report.increment()

            progress_report = task_group.group_progress_reports.filter(code="synclist.curate")
            progress_report.update(done=F("done") + len(synclist_ids))

    log.info(
        _("Finishing curating %s synclist repos based on %s update"),
        synclist_qs.count(),
        upstream_repository,
    )

    task_group.finish()


def curate_synclist_repository_batch(synclist_pks, **kwargs):
    for synclist_pk in synclist_pks:
        curate_synclist_repository(synclist_pk=synclist_pk)


def curate_synclist_repository(synclist_pk, **kwargs):
    """Update a synclist repo based on it's policy and spec.

    Update a curated synclist repo to use the latest versions from
    upstream as specified by the synclist's policy and it's collections
    and namespaces fields.

    This is intended to work on one synclist and synclist repo at a time.
    """

    synclist = models.SyncList.objects.get(pk=synclist_pk)

    upstream_repository = synclist.upstream_repository
    latest_upstream = upstream_repository.latest_version()

    log.info(
        _('Applying synclist "%s" with policy=%s to curate repo "%s" from upstream repo "%s"'),
        synclist.name,
        synclist.policy,
        synclist.repository.name,
        upstream_repository.name,
    )

    namespaces = synclist.namespaces.filter().values_list("name", flat=True)

    # include adds only highest version and exclude removes all versions of the collection.
    is_highest_query_param = {}
    if synclist.policy == 'include':
        is_highest_query_param = {"is_highest": True}

    collection_versions = CollectionVersion.objects.filter(
        Q(
            repositories=synclist.upstream_repository,
            collection__namespace__in=namespaces,
            **is_highest_query_param
        )
        | Q(
            collection__in=synclist.collections.all(),
            repositories=synclist.repository,
            **is_highest_query_param
        )
    )

    if synclist.policy == "exclude":
        task_kwargs = {
            "base_version_pk": str(latest_upstream.pk),
            "repository_pk": str(synclist.repository.pk),
            "add_content_units": [],
            "remove_content_units": collection_versions,
        }

    elif synclist.policy == "include":
        task_kwargs = {
            "repository_pk": str(synclist.repository.pk),
            "add_content_units": collection_versions,
            "remove_content_units": ["*"],
        }
    else:
        raise RuntimeError(_("Unexpected synclist policy {}").format(synclist.policy))

    add_and_remove(**task_kwargs)



import logging
from rest_framework import status
from django.test import override_settings

from galaxy_ng.app.models import auth as auth_models
from galaxy_ng.app import models
from pulp_container.app import models as container_models

from galaxy_ng.app.constants import DeploymentMode

from .base import BaseTestCase, get_current_ui_url

log = logging.getLogger(__name__)


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestContainerRemote(BaseTestCase):
    def setUp(self):
        super().setUp()
        permissions = [
            # change containers
            'container.namespace_change_containerdistribution',
            # create containers
            'container.add_containernamespace',
        ]
        self.container_user = auth_models.User.objects.create(username='container_user')
        self.admin = auth_models.User.objects.create(username='admin', is_superuser=True)
        self.regular_user = auth_models.User.objects.create(username='hacker')
        self.container_user.save()
        self.registry = models.ContainerRegistryRemote.objects.create(
            name="Test Registry",
            url="quay.io",
            proxy_url="example.com",
            username="foo",
            password="bar"
        )
        self.registry2 = models.ContainerRegistryRemote.objects.create(
            name="Test registry 2",
            url="registry.redhat.io",
        )
        self.remote_list_url = get_current_ui_url('execution-environments-remote-list')
        self.container_group = self._create_group(
            "",
            "container_group",
            users=[self.container_user, ],
            perms=permissions
        )

    def _create_remote(self, user, name, registry_pk, **kwargs):
        self.client.force_authenticate(user=user)
        return self.client.post(
            self.remote_list_url,
            {"name": name, "upstream_name": "foo", "registry": registry_pk, **kwargs},
            format='json'
        )

    def _update_remote(self, remote_pk, user, payload):
        self.client.force_authenticate(user=user)
        return self.client.put(
            f"{self.remote_list_url}{remote_pk}/",
            payload,
            format='json'
        )

    def _error_has_message(self, response, message, expected_errors=None, expected_field=None):
        """
        Checks that:
            1. the response object has an errors field
            2. that the given message matches one of the fields
            4. (optional) that the error matching the message is from the expected field
            3. (optional) that the number of errors is equal to expected_errors
        """
        errors = response.data.get('errors', None)
        self.assertTrue(errors is not None)

        if expected_errors:
            self.assertEqual(len(errors), expected_errors)

        for error in errors:
            if message.lower() in error["detail"].lower():
                if expected_field:
                    self.assertEqual(error['source']['parameter'], expected_field)
                return True

        return False

    def test_remote_creation(self):
        # test unprivileged users can't create
        response = self._create_remote(self.regular_user, 'remote/remote_repo2', self.registry.pk)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self._create_remote(self.container_user, 'remote/remote_repo', self.registry.pk)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # test values are copied over from registy
        remote_obj = container_models.ContainerRemote.objects.get(pk=response.data['pulp_id'])
        self.assertEqual(remote_obj.url, self.registry.url)
        self.assertEqual(remote_obj.proxy_url, self.registry.proxy_url)
        self.assertEqual(remote_obj.password, self.registry.password)
        self.assertEqual(remote_obj.registry.registry.pk, self.registry.pk)

        # test associated objects are created
        self.assertEqual(
            container_models.ContainerRepository.objects.filter(remote=remote_obj).count(), 1)
        repo = container_models.ContainerRepository.objects.filter(remote=remote_obj).first()

        self.assertEqual(repo.name, remote_obj.name)

        self.assertEqual(container_models.Distribution.objects.filter(repository=repo).count(), 1)
        distro = container_models.Distribution.objects.filter(repository=repo).first()
        self.assertEqual(distro.name, remote_obj.name)

        # test duplicate names not allowed
        bad_response = self._create_remote(
            self.container_user, 'remote/remote_repo', self.registry.pk)
        self.assertEqual(bad_response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue(
            self._error_has_message(bad_response, "must be unique", expected_field="name"))

    def test_remote_update(self):
        response = self._create_remote(self.admin, 'remote/remote_repo', self.registry.pk)
        obj_data = response.data
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        remote_pk = response.data['pulp_id']

        # test unpriviliged users can't update
        response = self._update_remote(
            remote_pk, self.regular_user, {**obj_data, "upstream_name": "streamy"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # test priviledged users can update
        response = self._update_remote(
            remote_pk,
            self.container_user,
            {**obj_data, "upstream_name": "streamy", "registry": self.registry2.pk}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # test values are copied over from registy
        remote_obj = container_models.ContainerRemote.objects.get(pk=remote_pk)
        self.assertEqual(remote_obj.url, self.registry2.url)
        self.assertEqual(remote_obj.proxy_url, self.registry2.proxy_url)
        self.assertEqual(remote_obj.password, self.registry2.password)
        self.assertEqual(remote_obj.registry.registry.pk, self.registry2.pk)

        # test changing name not allowed
        response = self._update_remote(
            remote_pk,
            self.container_user,
            {**obj_data, "name": "my_new_name"}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue(self._error_has_message(
            response, "cannot be changed", expected_field="name"))

    def test_remote_update_after_registry_deletion(self):
        """Ensure that a remote can be updated after the registry it's associated with is deleted

        1. create a registry
        2. create a remote in the registry
        3. delete the registry
        4. attempt to update the remote to use a new registry
        """
        # 1
        registry = models.ContainerRegistryRemote.objects.create(
            name="Test registry 3",
            url="docker.io",
        )

        # 2
        response = self._create_remote(self.admin, 'remote3/remote_repo3', registry.pk)
        obj_data = response.data
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        remote_pk = response.data['pulp_id']

        # 3
        registry.delete()

        # connection has been deleted
        self.assertFalse(
            models.container.ContainerRegistryRepos.objects.filter(
                registry=registry,
                repository_remote__pk=remote_pk
            ).exists()
        )

        # 4
        new_registry = models.ContainerRegistryRemote.objects.create(
            name="Test registry 4",
            url="myregistry.io",
        )

        # test update after registry deletion
        response = self._update_remote(
            remote_pk,
            self.container_user,
            {**obj_data, "upstream_name": "streamy", "registry": new_registry.pk}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # connection has been re-created
        self.assertTrue(
            models.container.ContainerRegistryRepos.objects.filter(
                registry=new_registry,
                repository_remote__pk=remote_pk
            ).exists()
        )

    def test_validation(self):
        # test invalid registry
        response = self._create_remote(
            self.container_user,
            "hello/i/am/invalid",
            "e47d47a9-30ea-4dd8-bebf-908893f59880"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue(self._error_has_message(
            response, "registry does not exist", expected_field="registry"))
        self.assertTrue(self._error_has_message(
            response, "names can only contain alphanumeric", expected_field="name"))

    def test_excludes_default(self):
        response = self._create_remote(
            self.admin, 'remote/remote_repo', self.registry.pk, exclude_tags=["foo"])
        obj_data = response.data

        self.assertTrue("*-source" in obj_data["exclude_tags"])
        self.assertEqual(len(obj_data["exclude_tags"]), 2)

        # verify exclude source can be removed.
        response = self._update_remote(
            obj_data['pulp_id'],
            self.container_user,
            {**obj_data, "exclude_tags": ["foo"]}
        )

        obj_data = response.data

        self.assertTrue("*-source" not in obj_data["exclude_tags"])
        self.assertEqual(len(obj_data["exclude_tags"]), 1)



import re
import socket

from django.utils.translation import gettext_lazy as _

from rest_framework import serializers
from pulpcore.plugin import models as pulp_models

from collections import namedtuple
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection
from urllib3.connectionpool import HTTPConnectionPool

from galaxy_importer.schema import MAX_LENGTH_NAME, MAX_LENGTH_VERSION

CollectionFilename = namedtuple("CollectionFilename", ["namespace", "name", "version"])

LOCALHOST = "localhost"
FILENAME_REGEXP = re.compile(
    r"^(?P<namespace>\w+)-(?P<name>\w+)-" r"(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$"
)
VERSION_REGEXP = re.compile(r"""
^
(?P<major>0|[1-9][0-9]*)
\.
(?P<minor>0|[1-9][0-9]*)
\.
(?P<patch>0|[1-9][0-9]*)
(?:
    -(?P<pre>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*)
)?
(?:
    \+(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*)
)?
$
""", re.VERBOSE | re.ASCII)


def parse_collection_filename(filename):
    """
    Parses collection filename.

    Parses and validates collection filename. Returns CollectionFilename named tuple.
    Raises ValueError if filename is not a valid collection filename.
    """
    match = FILENAME_REGEXP.match(filename)

    if not match:
        msg = _("Invalid filename {filename}. Expected format: namespace-name-version.tar.gz")
        raise ValueError(msg.format(filename=filename))

    namespace, name, version = match.groups()

    match = VERSION_REGEXP.match(version)
    if not match:
        msg = _("Invalid version string {version} from filename {filename}. Expected semantic version format.") # noqa
        raise ValueError(msg.format(version=version, filename=filename))

    if len(namespace) > MAX_LENGTH_NAME:
        raise ValueError(_("Expected namespace to be max length of %s") % MAX_LENGTH_NAME)
    if len(name) > MAX_LENGTH_NAME:
        raise ValueError(_("Expected name to be max length of %s") % MAX_LENGTH_NAME)
    if len(version) > MAX_LENGTH_VERSION:
        raise ValueError(_("Expected version to be max length of %s") % MAX_LENGTH_VERSION)

    return CollectionFilename(namespace, name, version)


class SocketHTTPConnection(HTTPConnection):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__(LOCALHOST)

    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.socket_file)


class SocketHTTPConnectionPool(HTTPConnectionPool):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__(LOCALHOST)

    def _new_conn(self):
        return SocketHTTPConnection(self.socket_file)


class SocketHTTPAdapter(HTTPAdapter):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__()

    def get_connection(self, url, proxies=None):
        return SocketHTTPConnectionPool(self.socket_file)


def get_write_only_fields(serializer, obj, extra_data=None):
    """
    Returns a list of write only fields and whether or not their values are set
    so that clients can tell if they are overwriting an existing value.
    serializer: Serializer instance
    obj: model object being serialized
    extra_data: extra fields that might not be on obj. This is used when a write
        only field is not one of the fields in the underlying data model.
    """
    fields = []
    extra_data = extra_data or {}

    # returns false if field is "" or None
    def _is_set(field_name):
        if (field_name in extra_data):
            return bool(extra_data[field_name])
        else:
            return bool(getattr(obj, field_name))

    # There are two ways to set write_only. This checks both.

    # check for values that are set to write_only in Meta.extra_kwargs
    for field_name in serializer.Meta.extra_kwargs:
        if serializer.Meta.extra_kwargs[field_name].get('write_only', False):
            fields.append({"name": field_name, "is_set": _is_set(field_name)})

    # check for values that are set to write_only in fields
    serializer_fields = serializer.get_fields()
    for field_name in serializer_fields:
        if (serializer_fields[field_name].write_only):
            fields.append({"name": field_name, "is_set": _is_set(field_name)})

    return fields


class RemoteSyncTaskField(serializers.Field):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, read_only=True)

    def to_representation(self, remote):
        # Query the database for sync tasks that reserve the given remote's PK
        task = pulp_models.Task.objects.filter(
            reserved_resources_record__icontains=remote.pk,
            name__icontains="sync"
        ).order_by('-pulp_last_updated').first()

        if not task:
            return {}
        else:
            return {
                "task_id": task.pk,
                "state": task.state,
                "started_at": task.started_at,
                "finished_at": task.finished_at,
                "error": task.error
            }



from django.db import migrations


def fix_inbound_repos(apps, schema_editor):
    """Add missing init RepositoryVersion to inbound repos
    created in 0003_inbound_repo_per_namespace."""

    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')
    RepositoryVersion = apps.get_model('core', 'RepositoryVersion')
    db_alias = schema_editor.connection.alias

    repos_missing_init_version = AnsibleRepository.objects.using(db_alias).filter(
        name__startswith='inbound-',
        next_version=0,
    )

    for repo in repos_missing_init_version:
        RepositoryVersion.objects.using(db_alias).create(
            repository=repo, number=0, complete=True)
        repo.next_version = 1
        repo.save()


class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0008_rename_default_repo'),
    ]

    operations = [
        migrations.RunPython(fix_inbound_repos, elidable=True),
    ]



import random
from argparse import ArgumentError
from collections import defaultdict
from functools import lru_cache

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.urls.base import reverse
from rest_framework.test import APIClient

from galaxy_ng.app.constants import INBOUND_REPO_NAME_FORMAT
from galaxy_ng.app.models.auth import Group, User
from galaxy_ng.app.models.namespace import Namespace, create_inbound_repo
from galaxy_ng.tests.constants import TAGS, TEST_COLLECTION_CONFIGS

STRATEGIES = ["test", "faux"]


def valid_prefix(prefix):
    if prefix and prefix[0].isdigit():
        raise CommandError("Prefix cannot start with a number")
    return prefix


class Command(BaseCommand):
    """Django management command to feed the system with testing collections
    Those collections are meant for testing only.
    namespace defaults to 'test'
    """

    def add_arguments(self, parser):
        parser.add_argument("--namespace", help="namespace name")
        parser.add_argument(
            "--strategy",
            help="generation stategy to call",
            type=str,
            default="faux",
            required=False,
            choices=STRATEGIES,
        )
        parser.add_argument(
            "-n",
            "--ns",
            help="number of namespaces to generate",
            required=False,
            type=int,
            default=21,  # 21 is good for pagination testing
        )
        parser.add_argument(
            "-c",
            "--cols",
            help="number of collections to generate",
            required=False,
            type=int,
            default=21,  # 21 is good for pagination testing
        )
        parser.add_argument(
            "-p",
            "--prefix",
            help="prefix for namespaces and collections",
            required=False,
            type=valid_prefix,
            default="",
        )
        parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="don't ask confirmation",
            default=False,
            required=False,
        )

    def echo(self, message):
        self.stdout.write(self.style.SUCCESS(message))

    @property
    def generator(self):
        """Imports on demand, avoids import error when dep not installed"""
        if not hasattr(self, "_generator"):
            from orionutils import generator  # noqa

            self._generator = generator
        return self._generator

    @property
    def faux(self):
        """Imports on demand, avoids import error when dep not installed"""
        if not hasattr(self, "_faux"):
            import fauxfactory  # noqa

            self._faux = fauxfactory
        return self._faux

    def handle(self, *args, **options):
        """Handle command"""

        if options.get("strategy") == "test":
            total = len(TEST_COLLECTION_CONFIGS)
        else:
            total = options["cols"] * options["ns"]

        if not options["yes"]:
            confirm = input(
                f"You are about to populate the system with {total} test collections.\n"
                "This operation cannot be undone\n"
                "Proceed? (Y/n)"
            ).lower()
            while 1:
                if confirm not in ("y", "n", "yes", "no"):
                    confirm = input('Please enter either "y/yes" or "n/no": ')
                    continue
                if confirm in ("y", "yes"):
                    break
                else:
                    self.echo("Process canceled.")
                    return

        self.client = APIClient()
        self.admin_user = User.objects.get(username="admin")
        self.client.force_authenticate(user=self.admin_user)
        self.prefix = options["prefix"].lower()
        self.strategy = options.get("strategy")
        getattr(self, f"strategy_{self.strategy}", self.raise_for_strategy)(*args, **options)

    def upload_collection(self, filename):
        collection_upload_url = reverse(
            "galaxy:api:v3:default-content:collection-artifact-upload",
        )
        return self.client.post(collection_upload_url, {"file": open(filename, "rb")})

    def raise_for_strategy(self, *args, **kwargs):
        """No strategy, no action"""
        raise ArgumentError("Selected strategy doen't exist")

    @lru_cache(maxsize=None)
    def create_namespace(self, namespace_name, groups=None):
        namespace_name = f"{self.prefix}{namespace_name}".lower()
        np, _ = Namespace.objects.get_or_create(name=namespace_name)
        if not np.groups:
            if groups:
                np.groups = groups
            else:
                pe = Group.objects.get(name="system:partner-engineers")
                np.groups = {
                    pe: [
                        "galaxy.upload_to_namespace",
                        "galaxy.change_namespace",
                        "galaxy.delete_namespace",
                    ]
                }
            np.save()
        self.echo(f"Created namespace {namespace_name}")
        self.echo(reverse("galaxy:api:v3:namespaces-detail", kwargs={"name": namespace_name}))

        return np

    def apply_prefix(self, config):
        """Apply a prefix to a config"""
        if self.prefix and self.prefix != "":
            for key in ["namespace", "name"]:
                if not config[key].startswith(self.prefix):
                    config[key] = f"{self.prefix}{config[key]}".lower()
        return config

    def build_collection(self, template, config):
        """Build a collection from a template and a config"""
        return self.generator.build_collection(template, config=config)

    def gen_version(self):
        x = random.randint(0, 3)
        y = random.randint(1, 9)
        z = random.randint(0, 9)
        return f"{x}.{y}.{z}"

    def apply_dep_prefix(self, dep, np):
        _, col = dep.split(".")
        collection_name = f"{self.prefix}{col}"
        return f"{np.name}.{collection_name}"

    def strategy_test(self, *args, **options):
        """Use constants.TEST_COLLECTION_CONFIGS to generate collections"""
        for config in TEST_COLLECTION_CONFIGS:
            namespace_name = config["namespace"]
            create_inbound_repo(namespace_name)
            np = self.create_namespace(namespace_name)
            config["namespace"] = np.name

            if settings.GALAXY_REQUIRE_CONTENT_APPROVAL:
                distro = INBOUND_REPO_NAME_FORMAT.format(namespace_name=np.name)
            else:
                distro = "published"

            _new_deps = {}
            if config.get("dependencies"):
                _new_deps = {
                    self.apply_dep_prefix(key, np): value
                    for key, value in config["dependencies"].items()
                }

            if _new_deps:
                config["dependencies"] = _new_deps

            config = self.apply_prefix(config)
            collection = self.build_collection("skeleton", config=config)
            response = self.upload_collection(collection.filename)

            self.echo(response.status_code)
            self.echo(response.data)
            self.echo(f"Collection '{collection.name}' created")
            if config.get("dependencies"):
                self.echo(f"Dependencies: {collection.name} depends on {config['dependencies']}")
            else:
                self.echo("No dependencies")
            self.echo(
                reverse(
                    "galaxy:api:content:v3:collection-versions-detail",
                    kwargs={
                        "path": distro,
                        "namespace": np.name,
                        "name": collection.name,
                        "version": collection.version,
                    },
                )
            )

    def strategy_faux(self, *args, **options):
        """Use fauxfactory to generate"""
        _configs = defaultdict(list)
        _dependency_pool = []
        for _ in range(options["ns"]):
            namespace_name = self.faux.gen_string(
                "alpha",
                8,
                validator=lambda v: v not in _configs,
                tries=100,
                default=self.faux.gen_string("alpha", 8),
            ).lower()
            create_inbound_repo(namespace_name)
            np = self.create_namespace(namespace_name)
            _collections = []

            if settings.GALAXY_REQUIRE_CONTENT_APPROVAL:
                distro = INBOUND_REPO_NAME_FORMAT.format(namespace_name=namespace_name)
            else:
                distro = "published"

            for i in range(options["cols"]):
                config = {
                    "name": self.faux.gen_string(
                        "alpha",
                        10,
                        validator=lambda v: v not in _collections,
                        tries=100,
                        default=self.faux.gen_string("alpha", 10),
                    ).lower(),
                    "description": self.faux.gen_iplum(words=8),
                    "namespace": np.name,
                    "version": self.gen_version(),
                    "tags": random.sample(TAGS, 3),
                }

                if i > 2:
                    deps = random.sample(_dependency_pool, random.randint(1, 3))
                    config["dependencies"] = dict(deps)

                _collections.append(config["name"])
                _configs[namespace_name].append(config)

                config = self.apply_prefix(config)
                collection = self.build_collection("skeleton", config=config)
                response = self.upload_collection(collection.filename)
                self.echo(response.status_code)
                self.echo(response.data)
                self.echo(f"Collection '{collection.name}' created")
                if config.get("dependencies"):
                    self.echo(
                        f"Dependencies: {collection.name} depends on {config['dependencies']}"
                    )
                else:
                    self.echo("No dependencies")
                self.echo(
                    reverse(
                        "galaxy:api:content:v3:collection-versions-detail",
                        kwargs={
                            "path": distro,
                            "namespace": np.name,
                            "name": config["name"],
                            "version": config["version"],
                        },
                    )
                )

                if i <= 2:
                    _config = config.copy()
                    x, y, z = config["version"].split(".")
                    _config["version"] = f"{x}.{y}.{int(z) + 1}"
                    _config = self.apply_prefix(_config)
                    _collection = self.build_collection("skeleton", config=_config)
                    _response = self.upload_collection(_collection.filename)
                    self.echo(_response.status_code)
                    self.echo(_response.data)
                    self.echo(f"A new version for '{_collection.name}' created")
                    self.echo(
                        reverse(
                            "galaxy:api:content:v3:collection-versions-detail",
                            kwargs={
                                "path": distro,
                                "namespace": np.name,
                                "name": _config["name"],
                                "version": _config["version"],
                            },
                        )
                    )

                    # add as a dependency to the next collection
                    spec = random.choice(["==", ">=", "<="])
                    _dependency_pool.append(
                        (f"{config['namespace']}.{config['name']}", f"{spec}{config['version']}")
                    )
                    _spec = random.choice(["==", ">=", "<="])
                    _dependency_pool.append(
                        (f"{config['namespace']}.{config['name']}", f"{_spec}{_config['version']}")
                    )



"""Tests that Collections hosted by Pulp can be installed by ansible-galaxy."""
from os import path
import subprocess
import tempfile

from galaxy_ng.tests.functional.utils import TestCaseUsingBindings
from galaxy_ng.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class InstallCollectionTestCase(TestCaseUsingBindings):
    """Test whether ansible-galaxy can install a Collection hosted by Pulp."""

    def install_collection(self, collection_name, repo_name="community"):
        """Install given collection."""

        # Preapare ansible.cfg for ansible-galaxy CLI
        self.update_ansible_cfg(repo_name)

        # In a temp dir, install a collection using ansible-galaxy CLI
        with tempfile.TemporaryDirectory() as temp_dir:
            cmd = "ansible-galaxy collection install -vvv {} -c -p {}".format(
                collection_name, temp_dir
            )

            directory = "{}/ansible_collections/{}".format(
                temp_dir, collection_name.replace(".", "/")
            )

            self.assertTrue(
                not path.exists(directory), "Directory {} already exists".format(directory)
            )

            subprocess.run(cmd.split())

            self.assertTrue(path.exists(directory), "Could not find directory {}".format(directory))

    def test_install_collection(self):
        """Test whether ansible-galaxy can install a Collection hosted by Pulp."""
        collection_name = "ansible.posix"
        # Sync the repository.
        self.sync_repo(f"collections:\n  - {collection_name}")
        self.install_collection(collection_name)



import logging

from galaxy_ng.app.api.v3.viewsets.namespace import NamespaceViewSet
from galaxy_ng.app.api.ui import versioning

log = logging.getLogger(__name__)


class NamespaceViewSet(NamespaceViewSet):
    versioning_class = versioning.UIVersioning



from rest_framework import versioning

from galaxy_ng.app import constants


# We can't use rest_framework.versioning.NamespaceVersioning directly because without
# setting the allowed versions, it will always return the first part of the namespace as
# the version, which will always resolve to "galaxy"
class UIVersioning(versioning.NamespaceVersioning):

    allowed_versions = list(constants.ALL_UI_API_VERSION.keys())



from django.urls.base import reverse
from django.test.utils import override_settings

from rest_framework.test import APIClient

from pulp_ansible.app.models import (
    AnsibleDistribution,
    AnsibleRepository,
    Collection,
    CollectionVersion,
)
from galaxy_ng.app import models
from galaxy_ng.app.constants import DeploymentMode

from .base import BaseTestCase


def _create_repo(name, **kwargs):
    repo = AnsibleRepository.objects.create(name=name, **kwargs)
    AnsibleDistribution.objects.create(
        name=name, base_path=name, repository=repo
    )
    return repo


def _get_create_version_in_repo(namespace, collection, version, repo):
    collection_version, _ = CollectionVersion.objects.get_or_create(
        namespace=namespace,
        name=collection.name,
        collection=collection,
        version=version,
    )
    qs = CollectionVersion.objects.filter(pk=collection_version.pk)
    with repo.new_version() as new_version:
        new_version.add_content(qs)


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class ViewOnlyTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.client = APIClient()

        self.namespace = models.Namespace.objects.create(name='view_only_namespace')
        self.collection = Collection.objects.create(
            namespace=self.namespace, name='view_only_collection'
        )
        self.repo = _create_repo(name='view_only_repo')

        _get_create_version_in_repo(
            self.namespace,
            self.collection,
            '1.1.1',
            self.repo
        )
        _get_create_version_in_repo(
            self.namespace,
            self.collection,
            '1.1.2',
            self.repo
        )

        self.collections_detail_url = reverse(
            'galaxy:api:content:v3:collections-detail',
            kwargs={
                'path': self.repo.name,
                'namespace': self.namespace.name,
                'name': self.collection.name
            }
        )

        self.ns_url = reverse('galaxy:api:v3:namespaces-list')

        self.ns_detail_url = reverse(
            'galaxy:api:v3:namespaces-detail',
            kwargs={
                "name": self.namespace.name
            }
        )

    def test_unauthenticated_access_to_collections(self):
        response = self.client.get(self.collections_detail_url)
        self.assertEqual(response.data['errors'][0]['status'], '403')
        with self.settings(GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS=True):
            response = self.client.get(self.collections_detail_url)
            self.assertEqual(response.data['name'], self.collection.name)
            self.assertEqual(response.data['namespace'], self.collection.namespace.name)
            self.assertEqual(response.data['deprecated'], False)
            self.assertEqual(
                response.data['highest_version']['version'], '1.1.2'
            )

    def test_unauthenticated_access_to_namespace(self):
        response = self.client.get(self.ns_detail_url)
        self.assertEqual(response.data['errors'][0]['status'], '403')
        with self.settings(GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS=True):
            response = self.client.get(self.ns_detail_url)
            self.assertEqual(response.data['name'], self.namespace.name)



from django.conf import settings
from django.contrib.auth import password_validation
from django.utils.translation import gettext_lazy as _
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field

from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from galaxy_ng.app.models import auth as auth_models


class GroupSerializer(serializers.ModelSerializer):

    class Meta:
        model = auth_models.Group
        fields = (
            'id',
            'name'
        )


class UserSerializer(serializers.ModelSerializer):
    auth_provider = serializers.SerializerMethodField()

    class Meta:
        model = auth_models.User
        fields = (
            'id',
            'username',
            'first_name',
            'last_name',
            'email',
            'groups',
            'password',
            'date_joined',
            'is_superuser',
            'auth_provider',
        )
        extra_kwargs = {
            'date_joined': {'read_only': True},
            'password': {'write_only': True, 'allow_blank': True, 'required': False}
        }

    def get_auth_provider(self, user):
        if hasattr(user, 'social_auth') and user.social_auth.all():
            providers = []
            for social in user.social_auth.all():
                providers.append(social.provider)
            return providers

        return ['django']

    def validate_password(self, password):
        if password:
            password_validation.validate_password(password)
            return password

    def validate_groups(self, groups):
        request_user = self.context['request'].user

        group_set = set(groups)
        instance_group_set = set()
        if self.instance:
            instance_group_set = set(list(self.instance.groups.all()))

        group_difference = instance_group_set.symmetric_difference(group_set)

        if not request_user.has_perm('galaxy.change_group'):
            authed_user_groups = request_user.groups.all()
            for g in group_difference:
                if not authed_user_groups.filter(pk=g.id).exists():
                    raise ValidationError(detail={
                        "groups": _("'galaxy.change_group' permission is required to change"
                                    " a users group that the requesting user is not in.")
                    })

        return groups

    def validate_is_superuser(self, data):
        request_user = self.context['request'].user

        # If the user is not a super user
        if not request_user.is_superuser:
            if self.instance:
                # Check if is_superuser is being modified, reject the request
                if self.instance.is_superuser != data:
                    raise ValidationError(detail={
                        "is_superuser": _("Must be a super user to grant super user permissions.")
                    })
            else:
                # if a new user is being created, reject the request if it is a super user
                if data:
                    raise ValidationError(detail={
                        "is_superuser": _("Must be a super user to grant super user permissions.")
                    })

        return data

    def _set_password(self, instance, data, updating):
        # password doesn't get set the same as other data, so delete it
        # before the serializer saves
        password = data.pop('password', None)
        if password:
            if updating:
                user = self.context['request'].user
                if not user.is_superuser and user.pk != instance.pk:
                    raise ValidationError(detail={
                        "password": _("Must be a super user to change another user's password.")
                    })

            instance.set_password(password)

        return instance

    def create(self, data):
        instance = super().create(data)
        instance = self._set_password(instance, data, updating=False)
        instance.save()
        return instance

    def update(self, instance, data):
        if instance.is_superuser and not self.context['request'].user.is_superuser:
            raise ValidationError(detail={
                "username": _("You do not have permissions to modify super users.")
            })

        instance = self._set_password(instance, data, updating=True)

        return super().update(instance, data)

    def to_representation(self, instance):
        representation = super().to_representation(instance)

        representation['groups'] = GroupSerializer(instance.groups.all(), many=True).data
        return representation

    def to_internal_value(self, data):
        groups = data.get('groups')
        if groups:
            group_ids = []
            for group in groups:
                group_filter = {}
                for field in group:
                    if field in ('id', 'name'):
                        group_filter[field] = group[field]
                try:
                    group = auth_models.Group.objects.get(**group_filter)
                    group_ids.append(group.id)
                except auth_models.Group.DoesNotExist:
                    raise ValidationError(detail={
                        'groups': _('Group name=%(name)s, id=%(id)s does not exist') % {
                            'name': group.get('name'), 'id': group.get('id')}
                    })
                except ValueError:
                    raise ValidationError(detail={'group': _('Invalid group name or ID')})
            data['groups'] = group_ids
        return super().to_internal_value(data)


class CurrentUserSerializer(UserSerializer):
    model_permissions = serializers.SerializerMethodField()

    class Meta(UserSerializer.Meta):
        model = auth_models.User
        fields = UserSerializer.Meta.fields + ('model_permissions', 'is_anonymous',)
        extra_kwargs = dict(
            groups={'read_only': True},
            **UserSerializer.Meta.extra_kwargs
        )

    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_model_permissions(self, obj):

        permissions = {
            # Collection Namespace
            "add_namespace": obj.has_perm('galaxy.add_namespace'),
            "upload_to_namespace": obj.has_perm('galaxy.upload_to_namespace'),
            "change_namespace": obj.has_perm('galaxy.change_namespace'),
            "delete_namespace": obj.has_perm('galaxy.delete_namespace'),

            # Collection Repository
            "move_collection": obj.has_perm('ansible.modify_ansible_repo_content'),
            "delete_collection": obj.has_perm('ansible.delete_collection'),

            # Collection Remote
            "add_remote": obj.has_perm('ansible.add_collectionremote'),
            "change_remote": obj.has_perm('ansible.change_collectionremote'),
            "delete_remote": obj.has_perm('ansible.delete_collectionremote'),

            # Collection Distribution
            "view_distribution": obj.has_perm('ansible.view_ansibledistribution'),
            "add_distribution": obj.has_perm('ansible.add_ansibledistribution'),
            "change_distribution": obj.has_perm('ansible.change_ansibledistribution'),
            "delete_distribution": obj.has_perm('ansible.delete_ansibledistribution'),

            # Container Namespace
            "add_containernamespace": obj.has_perm('container.add_containernamespace'),
            "change_containernamespace": obj.has_perm('container.change_containernamespace'),
            "delete_containernamespace": obj.has_perm('container.delete_containernamespace'),

            # Container Repository
            "add_containerrepository": obj.has_perm('container.add_containerrepository'),
            "change_containerrepository": obj.has_perm('container.change_containerrepository'),
            "delete_containerrepository": obj.has_perm('container.delete_containerrepository'),

            # Container Remote
            "add_containerregistry": obj.has_perm('galaxy.add_containerregistryremote'),
            "change_containerregistry": obj.has_perm('galaxy.change_containerregistryremote'),
            "delete_containerregistry": obj.has_perm('galaxy.delete_containerregistryremote'),

            # Container Distribution
            "add_containerdistribution": obj.has_perm('container.add_containerdistribution'),
            "change_containerdistribution": obj.has_perm('container.change_containerdistribution'),
            "delete_containerdistribution": obj.has_perm('container.delete_containerdistribution'),

            # Auth
            "view_user": obj.has_perm('galaxy.view_user'),
            "change_group": obj.has_perm('galaxy.change_group'),
            "view_group": obj.has_perm('galaxy.view_group'),
            "delete_user": False,
            "change_user": False,
            "add_user": False,
            "add_group": False,
            "delete_group": False,
        }

        if not settings.get("SOCIAL_AUTH_KEYCLOAK_KEY"):
            permissions["delete_user"] = obj.has_perm('galaxy.delete_user')
            permissions["change_user"] = obj.has_perm('galaxy.change_user')
            permissions["add_user"] = obj.has_perm('galaxy.add_user')
            permissions["add_group"] = obj.has_perm('galaxy.add_group')
            permissions["delete_group"] = obj.has_perm('galaxy.delete_group')

        return permissions



from django.contrib.auth import get_user_model
from django.core.management import BaseCommand
from django.core.management import CommandError

User = get_user_model()


class Command(BaseCommand):
    """
    Django management command for creating a user
    """

    help = 'Create a user'

    def add_arguments(self, parser):
        parser.add_argument('--user', metavar='USER', dest='username',
                            help='user name', required=True)

    def handle(self, *args, **options):
        username = options['username']
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            try:
                user = User.objects.create(username=username)
                user.save()
            except Exception as e:
                raise CommandError("Could not create user: {}".format(e))

            self.stdout.write("Created user '{}'".format(username))

        self.stdout.write("User '{}' already exists".format(username))



"""
Constants to be use by testing scripts and development management commands.
"""

TEST_COLLECTION_CONFIGS = [
    {"name": "a", "namespace": "test", "dependencies": {"test.b": "*"}},
    {"name": "b", "namespace": "test", "dependencies": {"test.c": "*"}},
    {"name": "c", "namespace": "test", "dependencies": {"test.d": "*"}},
    {"name": "d", "namespace": "test"},
    {"name": "e", "namespace": "test", "dependencies": {"test.f": "*", "test.g": "*"}},
    {"name": "f", "namespace": "test", "dependencies": {"test.h": "<=3.0.0"}},
    {"name": "g", "namespace": "test", "dependencies": {"test.d": "*"}},
    {"name": "h", "namespace": "test", "version": "1.0.0"},
    {"name": "h", "namespace": "test", "version": "2.0.0"},
    {"name": "h", "namespace": "test", "version": "3.0.0"},
    {"name": "h", "namespace": "test", "version": "4.0.0"},
    {"name": "h", "namespace": "test", "version": "5.0.0"},
]
"""
TEST_COLLECTION_CONFIGS Dependency Trees:

        A               E
        |             /   \
        B            F     G
        |            |     |
        C         H(1-3)   D
        |
        D

        H has 5 versions, no dependencies


The TEST_COLLECTION_CONFIGS are used to build collections with orionutils.

ex:

    from orionutils.generator import build_collection
    collection = build_collection("skeleton", config=TEST_COLLECTION_CONFIGS)
    print(collection.filename)

"""

TAGS = [
    "system",
    "database",
    "web",
    "middleware",
    "infra",
    "linux",
    "rhel",
    "fedora",
    "cloud",
    "monitoring",
    "windows",
    "nginx_config",
    "container",
    "docker",
    "development",
    "kubernetes",
    "java",
    "python",
    "rust",
    "proxy",
]




import logging

from django.urls import reverse
from pulp_ansible.app.models import AnsibleRepository, AnsibleDistribution
from rest_framework import status

from galaxy_ng.app.models import auth as auth_models
from galaxy_ng.app.models import Namespace
from galaxy_ng.app.api.v3.serializers import NamespaceSerializer
from galaxy_ng.app.constants import DeploymentMode, INBOUND_REPO_NAME_FORMAT

from .base import BaseTestCase

log = logging.getLogger(__name__)

# /api/automation-hub/v3/namespaces/
# galaxy_ng.app.api.v3.viewsets.namespace.NamespaceViewSet
# galaxy:api:v3:namespaces-list

# /api/automation-hub/v3/namespaces/<name>/
# galaxy_ng.app.api.v3.viewsets.namespace.NamespaceViewSet
# galaxy:api:v3:namespaces-detail


class TestV3NamespaceViewSet(BaseTestCase):
    deployment_mode = DeploymentMode.STANDALONE.value

    def setUp(self):
        super().setUp()
        self.admin_user = auth_models.User.objects.create(username='admin')
        self.pe_group = self._create_partner_engineer_group()
        self.admin_user.groups.add(self.pe_group)
        self.admin_user.save()

        self.regular_user = auth_models.User.objects.create(username='regular')

        self.ns_url = reverse('galaxy:api:v3:namespaces-list')

    def test_namespace_validation(self):
        ns_name = "unittestnamespace"
        ns1 = self._create_namespace(ns_name, groups=[self.pe_group])

        ns_detail_url = reverse('galaxy:api:v3:namespaces-detail', kwargs={"name": ns1.name})

        self.client.force_authenticate(user=self.admin_user)

        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            put_data = {
                "name": ns1.name,
                "groups": [],
                "company": "Super duper long name that nobody in their right "
                           "mind should really ever use for a name on anything anywhere.",
                "avatar_url": "not a url",
                "links": [
                    {
                        "name": "link name that is way way too long for anyone",
                        "url": "not a url",
                    },
                    {
                        "name": "link name that is way way too long for anyone",
                        "url": "https://www.valid-url.com/happyface.jpg",
                    }
                ]
            }
            expected_error_fields = {'company', 'avatar_url', 'links__url', 'links__name'}
            expected_error_codes = {'max_length', 'invalid'}

            response = self.client.put(ns_detail_url, put_data, format='json')

            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertEqual(len(response.data['errors']), 5)

            received_error_fields = set()
            received_error_codes = set()
            for err in response.data['errors']:
                received_error_codes.add(err['code'])
                received_error_fields.add(err['source']['parameter'])

            self.assertEqual(expected_error_codes, received_error_codes)
            self.assertEqual(expected_error_fields, received_error_fields)

    def test_namespace_list_empty(self):
        self.client.force_authenticate(user=self.admin_user)
        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            response = self.client.get(self.ns_url)
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data['data']
            log.debug("data: %s", data)
            self.assertEqual(len(data), Namespace.objects.all().count())

    def test_namespace_list(self):
        self.client.force_authenticate(user=self.admin_user)
        ns1_name = "unittestnamespace1"
        ns2_name = "unittestnamespace2"
        self._create_namespace(ns1_name, groups=[self.pe_group])
        self._create_namespace(ns2_name, groups=[self.pe_group])
        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            response = self.client.get(self.ns_url)
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data['data']
            log.debug("data: %s", data)
            self.assertEqual(len(data), Namespace.objects.all().count())

    def test_namespace_get(self):
        ns_name = "unittestnamespace"
        ns1 = self._create_namespace(ns_name, groups=[self.pe_group])

        ns_detail_url = reverse('galaxy:api:v3:namespaces-detail', kwargs={"name": ns1.name})

        self.client.force_authenticate(user=self.admin_user)

        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            response = self.client.get(ns_detail_url)
            log.debug('response: %s', response)

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.data
            log.debug('data: %s', data)

            namespace = NamespaceSerializer(data=data)
            log.debug('repr namespace serializer: %r', namespace)

            namespace.is_valid()
            log.debug('namespace.errors: %s', namespace.errors)

            log.debug('namespace.validated_data: %s', namespace.validated_data)

            self.assertEqual(data['name'], ns_name)
            self.assertEqual(data['name'], ns1.name)

            self.assertEqual(data["links"], [])
            self.assertIn("avatar_url", data)
            self.assertIn("company", data)
            self.assertIn("description", data)

            self.assertEqual(len(data['groups']), self.admin_user.groups.all().count())

    def test_namespace_api_creates_deletes_inbound_repo(self):
        self.client.force_authenticate(user=self.admin_user)
        ns1_name = "unittestnamespace1"
        repo_name = INBOUND_REPO_NAME_FORMAT.format(namespace_name=ns1_name)

        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            # Create namespace + repo
            response = self.client.post(
                self.ns_url,
                {
                    "name": ns1_name,
                    "groups": [
                        {
                            "id": self.pe_group.id,
                            "name": self.pe_group.name,
                            "object_permissions": [
                                'galaxy.upload_to_namespace',
                                'galaxy.change_namespace',
                                'galaxy.delete_namespace',
                            ]
                        },
                    ],
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
            self.assertEqual(1, len(AnsibleRepository.objects.filter(name=repo_name)))
            self.assertEqual(1, len(AnsibleDistribution.objects.filter(name=repo_name)))

            # List namespace
            response = self.client.get(self.ns_url)
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            namespaces = [item['name'] for item in response.data['data']]
            self.assertIn(ns1_name, namespaces)
            self.assertEqual(1, len(Namespace.objects.filter(name=ns1_name)))

            # Delete namespace + repo
            ns_detail_url = reverse('galaxy:api:v3:namespaces-detail', kwargs={"name": ns1_name})
            response = self.client.delete(ns_detail_url)
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
            self.assertEqual(0, len(AnsibleRepository.objects.filter(name=repo_name)))
            self.assertEqual(0, len(AnsibleDistribution.objects.filter(name=repo_name)))
            self.assertEqual(0, len(Namespace.objects.filter(name=ns1_name)))

    def test_unauthorized_user_cant_delete_namespace(self):
        ns1_name = "unittestnamespacefordeletion"
        repo_name = INBOUND_REPO_NAME_FORMAT.format(namespace_name=ns1_name)
        self._create_namespace(ns1_name, groups=[self.pe_group])

        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            self.client.force_authenticate(user=self.regular_user)
            ns_detail_url = reverse('galaxy:api:v3:namespaces-detail', kwargs={"name": ns1_name})
            response = self.client.delete(ns_detail_url)
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
            self.assertEqual(1, len(Namespace.objects.filter(name=ns1_name)))
            self.assertEqual(1, len(AnsibleRepository.objects.filter(name=repo_name)))
            self.assertEqual(1, len(AnsibleDistribution.objects.filter(name=repo_name)))

    def test_delete_namespace_no_error_if_no_repo_exist(self):
        ns2_name = "unittestnamespace2"
        self._create_namespace(ns2_name, groups=[self.pe_group])
        ns_detail_url = reverse('galaxy:api:v3:namespaces-detail', kwargs={"name": ns2_name})
        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            response = self.client.delete(ns_detail_url)
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_conflict_error_if_already_exists(self):
        """Ensure API returns 409-Conflict in case of existing namespace."""
        ns3_name = "unittestnamespace3"
        self._create_namespace(ns3_name, groups=[self.pe_group])
        ns_list_url = reverse('galaxy:api:v3:namespaces-list')

        self.client.force_authenticate(user=self.admin_user)

        with self.settings(GALAXY_DEPLOYMENT_MODE=self.deployment_mode):
            post_data = {
                "name": ns3_name,
                "company": "A company name",
                "email": "email@companyname.com",
                "description": "A testing namespace",
                "groups": [
                    {
                        "id": self.pe_group.id,
                        "name": self.pe_group.name,
                        "object_permissions": ["change_namespace", "upload_to_namespace"]
                    }
                ]
            }
            response = self.client.post(ns_list_url, post_data, format='json')
            log.debug('namespace response: %s', response.data)
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)



import logging
from typing import Any, Dict, Optional
from rest_framework import serializers
from rest_framework.reverse import reverse
from drf_spectacular.utils import extend_schema_field
from pulpcore.app.serializers import ProgressReportSerializer
from pulpcore.plugin.models import Task


log = logging.getLogger(__name__)


class TaskSerializer(serializers.ModelSerializer):
    created_at = serializers.DateTimeField(source='pulp_created')
    updated_at = serializers.DateTimeField(source='pulp_last_updated')
    worker = serializers.SerializerMethodField()
    progress_reports = ProgressReportSerializer(many=True, read_only=True)
    pulp_id = serializers.UUIDField(source='pk')
    name = serializers.CharField()
    state = serializers.CharField()
    started_at = serializers.DateTimeField()
    finished_at = serializers.DateTimeField()

    @extend_schema_field(Optional[Dict[str, Any]])
    def get_worker(self, obj):
        if obj.worker:
            return {
                'name': obj.worker.name,
                'missing': obj.worker.missing,
                'last_heartbeat': obj.worker.last_heartbeat,
            }

    class Meta:
        model = Task
        ref_name = "Task"
        fields = (
            'pulp_id',
            'name',
            'created_at',
            'updated_at',
            'finished_at',
            'started_at',
            'state',
            'error',
            'worker',
            'parent_task',
            'child_tasks',
            'progress_reports',
        )


class TaskSummarySerializer(TaskSerializer):
    """TaskSerializer but without detail fields.

    For use in /tasks/<str:pk>/ detail views."""
    href = serializers.SerializerMethodField()

    def get_href(self, obj) -> str:
        return reverse(
            'galaxy:api:v3:default-content:tasks-detail',
            kwargs={"pk": str(obj.pk)}
        )

    class Meta:
        model = Task
        fields = (
            'pulp_id',
            'name',
            'state',
            'started_at',
            'finished_at',
            'href',
        )



from django.conf import settings
from django.db import transaction
from guardian.shortcuts import get_groups_with_perms, assign_perm, remove_perm
from django_lifecycle import hook


class GroupModelPermissionsMixin:
    _groups = None

    @property
    def groups(self):
        return get_groups_with_perms(self, attach_perms=True)

    @groups.setter
    def groups(self, groups):
        self._set_groups(groups)

    @transaction.atomic
    def _set_groups(self, groups):
        # guardian doesn't allow adding permissions to objects that haven't been
        # saved. When creating new objects, save group data to _groups where it
        # can be picked up by the post save hook.
        if self._state.adding:
            self._groups = groups
        else:
            current_groups = get_groups_with_perms(self, attach_perms=True)
            for group in current_groups:
                for perm in current_groups[group]:
                    remove_perm(perm, group, self)

            for group in groups:
                for perm in groups[group]:
                    assign_perm(perm, group, self)

    @hook('after_save')
    def set_object_groups(self):
        if self._groups:
            self._set_groups(self._groups)


class UnauthenticatedCollectionAccessMixin:
    def unauthenticated_collection_access_enabled(self, request, view, action):
        return settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS



import string
import random

from pulpcore.client.galaxy_ng.exceptions import ApiException
from galaxy_ng.tests.functional.utils import TestCaseUsingBindings
from galaxy_ng.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class CreateNamespaceTestCase(TestCaseUsingBindings):
    """Test whether a namespace can be created."""

    def delete_namespace(self, namespace_name):
        # delete namespace
        # namespace_api does not support delete, so we can use the smash_client directly
        response = self.smash_client.delete(
            f"{self.galaxy_api_prefix}/v3/namespaces/{namespace_name}"
        )
        self.assertEqual(response.status_code, 204)

    def test_create_and_delete_namespace(self):
        # generate name formed by 10 random ascii lowercase letters
        random_name = ''.join(random.choices(string.ascii_lowercase, k=10))
        namespace_data = {"name": random_name, "groups": []}

        # create namespace
        namespace = self.namespace_api.create(namespace=namespace_data)
        self.assertEqual(namespace.name, random_name)

        # ensure namespace is available
        namespaces = self.namespace_api.list(limit=100)
        self.assertIn(namespace.name, [item.name for item in namespaces.data])

        # delete namespace
        self.delete_namespace(namespace.name)

        # ensure namespace is NO MORE available
        namespaces = self.namespace_api.list(limit=100)
        self.assertNotIn(namespace.name, [item.name for item in namespaces.data])

    def test_negative_create_namespace_with_invalid_name(self):
        # generate name formed by 10 random ascii lowercase letters
        random_name = ''.join(random.choices(string.ascii_lowercase, k=10))
        random_name = f"ABC-{random_name}-$@"
        namespace_data = {"name": random_name, "groups": []}

        # expect the namespace is not created because of invalid name
        with self.assertRaises(ApiException) as context:
            self.namespace_api.create(namespace=namespace_data)

        # Bad request
        assert context.exception.status == 400



from pulpcore.plugin import viewsets as pulp_viewsets
from rest_framework import mixins

from galaxy_ng.app import models
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api.ui import serializers

# This file is necesary to prevent the DRF web API browser from breaking on all of the
# pulp/api/v3/repositories/ endpoints.

# Pulp expects models it manages to come with a viewset attached to them. Since the
# ContainerRegistryRemote model is a pulp remote, pulp expects it to have a pulp
# viewset attached to it. Pulp associates viewsets with models by looking in
# <plugin_name>.app.viewsets. Since galaxy_ng stores it's viewsets in a different
# module, this file is necesary for pulp tbe able to associate the ContainerRegistryRemote
# model to a viewset.

# Without this viewset defined here, pulp get's confused when it tries to auto generate
# the form on the repositories endpoint because it tries to provide a dropdown
# on the remote field and can't find a viewset name for galaxy's ContainerRegistryRemote model.


class ContainerRegistryRemoteViewSet(pulp_viewsets.NamedModelViewSet, mixins.RetrieveModelMixin):
    queryset = models.ContainerRegistryRemote.objects.all()
    serializer_class = serializers.ContainerRegistryRemoteSerializer
    permission_classes = [access_policy.ContainerRegistryRemoteAccessPolicy]
    endpoint_name = "execution-environments-registry-detail"


class ContainerDistributionViewSet(
    pulp_viewsets.NamedModelViewSet,
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
):
    queryset = models.ContainerDistribution.objects.all()
    serializer_class = serializers.ContainerRepositorySerializer
    permission_classes = [access_policy.ContainerRepositoryAccessPolicy]
    endpoint_name = "container"



import logging

from django.shortcuts import get_object_or_404
from guardian.shortcuts import get_objects_for_user

from rest_framework.decorators import action

from pulpcore.plugin.viewsets import OperationPostponedResponse
from pulpcore.plugin.tasking import dispatch

from galaxy_ng.app import models
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.tasks import curate_synclist_repository

from .synclist import SyncListViewSet


log = logging.getLogger(__name__)


class MySyncListViewSet(SyncListViewSet):
    permission_classes = [
        access_policy.MySyncListAccessPolicy,
    ]

    def get_queryset(self):
        """
        Returns all synclists for the user.
        """
        return get_objects_for_user(
            self.request.user,
            "galaxy.change_synclist",
            any_perm=True,
            accept_global_perms=False,
            klass=models.SyncList,
        )

    @action(detail=True, methods=["post"])
    def curate(self, request, pk):
        synclist = get_object_or_404(models.SyncList, pk=pk)
        synclist_task = dispatch(
            curate_synclist_repository,
            resources=[synclist.repository],
            args=(pk, )
        )

        log.debug("synclist_task: %s", synclist_task)

        return OperationPostponedResponse(synclist_task, request)



import logging
from rest_framework import status
from django.urls import reverse
from django.test import override_settings
from pulp_ansible.app.models import (
    AnsibleDistribution,
    AnsibleRepository,
    CollectionRemote
)
from galaxy_ng.app.constants import DeploymentMode
from .base import BaseTestCase

log = logging.getLogger(__name__)


def _create_repo(name, **kwargs):
    repo = AnsibleRepository.objects.create(name=name, **kwargs)
    AnsibleDistribution.objects.create(name=name, base_path=name, repository=repo)
    return repo


def _create_remote(name, url, **kwargs):
    return CollectionRemote.objects.create(
        name=name, url=url, **kwargs
    )


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestUiSyncConfigViewSet(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.admin_user = self._create_user("admin")
        self.pe_group = self._create_partner_engineer_group()
        self.admin_user.groups.add(self.pe_group)
        self.admin_user.save()

        # Remotes are created by data migration
        self.certified_remote = CollectionRemote.objects.get(name='rh-certified')
        self.community_remote = CollectionRemote.objects.get(name='community')

    def build_config_url(self, path):
        return reverse('galaxy:api:content:v3:sync-config', kwargs={'path': path})

    def build_sync_url(self, path):
        return reverse('galaxy:api:content:v3:sync', kwargs={'path': path})

    def test_positive_get_config_sync_for_certified(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(self.build_config_url(self.certified_remote.name))
        log.debug('test_positive_get_config_sync_for_certified')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)
        self.assertEqual(response.data['name'], self.certified_remote.name)
        self.assertEqual(response.data['url'], self.certified_remote.url)
        self.assertEqual(response.data['requirements_file'], None)

    def test_positive_get_config_sync_for_community(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(self.build_config_url(self.community_remote.name))
        log.debug('test_positive_get_config_sync_for_community')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)
        self.assertEqual(response.data['name'], self.community_remote.name)
        self.assertEqual(response.data['url'], self.community_remote.url)
        self.assertEqual(
            response.data['requirements_file'],
            self.community_remote.requirements_file
        )

    def test_positive_update_certified_repo_data(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.put(
            self.build_config_url(self.certified_remote.name),
            {
                "auth_url": "https://auth.com",
                "token": "TEST",
                "name": "rh-certified",
                "policy": "immediate",
                "requirements_file": None,
                "url": "https://updated.url.com/",
            },
            format='json'
        )
        log.debug('test_positive_update_certified_repo_data')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        updated = self.client.get(self.build_config_url(self.certified_remote.name))
        self.assertEqual(updated.data["auth_url"], "https://auth.com")
        self.assertEqual(updated.data["url"], "https://updated.url.com/")
        self.assertIsNone(updated.data["requirements_file"])

    def test_negative_update_community_repo_data_without_requirements_file(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.put(
            self.build_config_url(self.community_remote.name),
            {
                "auth_url": "https://auth.com",
                "name": "community",
                "policy": "immediate",
                "requirements_file": None,
                "url": "https://galaxy.ansible.com/v3/collections",
            },
            format='json'
        )
        log.debug('test_negative_update_community_repo_data_without_requirements_file')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn(
            'Syncing content from community domains without specifying a '
            'requirements file is not allowed.',
            str(response.data['errors'])
        )

    def test_positive_update_community_repo_data_with_requirements_file(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.put(
            self.build_config_url(self.community_remote.name),
            {
                "auth_url": "https://auth.com",
                "token": "TEST",
                "name": "community",
                "policy": "immediate",
                "requirements_file": (
                    "collections:\n"
                    "  - name: foo.bar\n"
                    "    server: https://foobar.content.com\n"
                    "    api_key: s3cr3tk3y\n"
                ),
                "url": "https://galaxy.ansible.com/v3/collections/",
            },
            format='json'
        )

        log.debug('test_positive_update_community_repo_data_with_requirements_file')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('foobar.content.com', response.data['requirements_file'])

    def test_positive_syncing_returns_a_task_id(self):
        self.client.force_authenticate(user=self.admin_user)
        response = self.client.post(self.build_sync_url(self.certified_remote.name))
        log.debug('test_positive_syncing_returns_a_task_id')
        log.debug('response: %s', response)
        log.debug('response.data: %s', response.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('task', response.data)

    def test_sensitive_fields_are_not_exposed(self):
        self.client.force_authenticate(user=self.admin_user)
        api_url = self.build_config_url(self.certified_remote.name)
        response = self.client.get(api_url)
        self.assertNotIn('password', response.data)
        self.assertNotIn('token', response.data)
        self.assertNotIn('proxy_password', response.data)

    def test_write_only_fields(self):
        self.client.force_authenticate(user=self.admin_user)
        api_url = self.build_config_url(self.certified_remote.name)
        write_only_fields = [
            'client_key',
            'token',
            'password',
            'proxy_password',
        ]

        # note, proxy (url, username and password) are required together
        request_data = {
            "url": self.certified_remote.url,
            "proxy_url": "https://example.com",
            "proxy_username": "bob",
            "proxy_password": "1234",
        }

        for field in write_only_fields:
            request_data[field] = "value_is_set"

        self.client.put(api_url, request_data, format='json')

        write_only = self.client.get(api_url).data['write_only_fields']
        response_names = set()
        # Check that all write only fields are set
        for field in write_only:
            self.assertEqual(field['is_set'], True)

            # unset all write only fields
            request_data[field['name']] = None
            response_names.add(field['name'])

        # proxy username and password can only be specified together
        request_data["proxy_username"] = None
        self.assertEqual(set(write_only_fields), response_names)
        response = self.client.put(api_url, request_data, format='json')
        self.assertEqual(response.status_code, 200)

        response = self.client.get(api_url)
        self.assertEqual(response.status_code, 200)

        # Check that proxy_username is unset
        self.assertIsNone(response.data['proxy_username'])

        # Check that all write only fields are unset
        write_only = response.data['write_only_fields']
        for field in write_only:
            self.assertEqual(field['is_set'], False)
            request_data[field['name']] = ""

        self.client.put(api_url, request_data, format='json')

        write_only = self.client.get(api_url).data['write_only_fields']
        for field in write_only:
            self.assertEqual(field['is_set'], False)

    def test_proxy_fields(self):
        self.client.force_authenticate(user=self.admin_user)

        # ensure proxy_url is blank
        api_url = self.build_config_url(self.certified_remote.name)
        response = self.client.get(api_url)
        self.assertIsNone(response.data['proxy_url'])

        data = {'name': response.data['name'], 'url': response.data['url']}

        # PUT proxy url without auth
        self.client.put(api_url, {'proxy_url': 'http://proxy.com:4242', **data}, format='json')
        response = self.client.get(api_url)
        self.assertEqual(response.data['proxy_url'], 'http://proxy.com:4242')
        self.assertNotIn('proxy_password', response.data)
        self.assertIn('proxy_username', response.data)
        instance = CollectionRemote.objects.get(pk=response.data['pk'])
        self.assertEqual(instance.proxy_url, 'http://proxy.com:4242')

        # PUT proxy url with username and password
        self.client.put(
            api_url,
            {
                'proxy_url': 'http://proxy.com:4242',
                'proxy_username': 'User1',
                'proxy_password': 'MyPrecious42',
                **data
            },
            format='json'
        )
        response = self.client.get(api_url)
        self.assertEqual(response.data['proxy_url'], 'http://proxy.com:4242')
        self.assertNotIn('proxy_password', response.data)
        self.assertIn('proxy_username', response.data)
        instance = CollectionRemote.objects.get(pk=response.data['pk'])
        self.assertEqual(instance.proxy_url, 'http://proxy.com:4242')
        self.assertEqual(instance.proxy_username, 'User1')
        self.assertEqual(instance.proxy_password, 'MyPrecious42')

        # Edit url using IP
        self.client.put(api_url, {'proxy_url': 'http://192.168.0.42:4242', **data}, format='json')
        response = self.client.get(api_url)
        self.assertEqual(response.data['proxy_url'], 'http://192.168.0.42:4242')
        self.assertNotIn('proxy_password', response.data)
        self.assertIn('proxy_username', response.data)
        instance = CollectionRemote.objects.get(pk=response.data['pk'])
        self.assertEqual(instance.proxy_url, 'http://192.168.0.42:4242')

        # Edit url
        self.client.put(api_url, {'proxy_url': 'http://proxy2.com:4242', **data}, format='json')
        response = self.client.get(api_url)
        self.assertEqual(response.data['proxy_url'], 'http://proxy2.com:4242')
        self.assertNotIn('proxy_password', response.data)
        self.assertIn('proxy_username', response.data)
        instance = CollectionRemote.objects.get(pk=response.data['pk'])
        self.assertEqual(instance.proxy_url, 'http://proxy2.com:4242')



import logging

from galaxy_ng.app import models
from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api.ui import serializers


log = logging.getLogger(__name__)


class SyncListViewSet(api_base.ModelViewSet):
    queryset = models.SyncList.objects.all()
    serializer_class = serializers.SyncListSerializer
    permission_classes = [access_policy.MySyncListAccessPolicy]



from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.exceptions import ValidationError

from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema

from pulp_container.app import models as pulp_models
from pulpcore.plugin.viewsets import (
    OperationPostponedResponse,
)
from pulpcore.plugin.serializers import AsyncOperationResponseSerializer
from pulpcore.plugin.tasking import dispatch

from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app import models, tasks


class ContainerSyncRemoteView(api_base.APIView):
    permission_classes = [access_policy.ContainerRemoteAccessPolicy]
    action = 'sync'

    def get_object(self):
        """Object is a ContainerRemote instance"""
        distro = self.get_distribution()
        remote = distro.repository.remote.cast()
        return remote

    def get_distribution(self):
        distro_path = self.kwargs['base_path']
        distro = get_object_or_404(pulp_models.ContainerDistribution, base_path=distro_path)

        if not distro.repository or not distro.repository.remote:
            raise ValidationError(
                detail={'remote': _('The %s distribution does not have'
                                    ' any remotes associated with it.') % distro_path})

        return distro

    def post(self, request: Request, *args, **kwargs) -> Response:
        distro_path = kwargs['base_path']
        distro = self.get_distribution()
        remote = distro.repository.remote.cast()

        try:
            registry = remote.registry.registry
        except models.ContainerRegistryRepos.repository_remote.RelatedObjectDoesNotExist:
            raise ValidationError(
                detail={'remote': _('The %s remote does not have'
                                    ' any registries associated with it.') % distro_path}
            )

        result = tasks.launch_container_remote_sync(remote, registry, distro.repository)
        return OperationPostponedResponse(result, request)


class ContainerSyncRegistryView(api_base.APIView):
    permission_classes = [access_policy.ContainerRegistryRemoteAccessPolicy]
    action = 'sync'

    @extend_schema(
        description="Trigger an asynchronous sync task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def post(self, request: Request, *args, **kwargs) -> Response:
        registry = get_object_or_404(models.ContainerRegistryRemote, pk=kwargs['pk'])

        result = dispatch(
            tasks.sync_all_repos_in_registry,
            kwargs={
                "registry_pk": str(registry.pk),
            },
            exclusive_resources=[registry]
        )
        return OperationPostponedResponse(result, request)



import logging

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_access_policy import AccessPolicy
from rest_framework.exceptions import NotFound

from pulp_container.app.access_policy import NamespacedAccessPolicyMixin
from pulp_container.app import models as container_models

from galaxy_ng.app import models
from galaxy_ng.app.access_control.mixins import UnauthenticatedCollectionAccessMixin

log = logging.getLogger(__name__)


class AccessPolicyBase(AccessPolicy):

    _STATEMENTS = None

    @property
    def galaxy_statements(self):
        """Lazily import the galaxy_statements from the statements file."""
        if self._STATEMENTS is None:
            # import here to avoid working outside django/dynaconf settings context
            from galaxy_ng.app.access_control.statements import STANDALONE_STATEMENTS  # noqa
            from galaxy_ng.app.access_control.statements import INSIGHTS_STATEMENTS  # noqa
            self._STATEMENTS = {
                'insights': INSIGHTS_STATEMENTS,
                'standalone': STANDALONE_STATEMENTS
            }
        return self._STATEMENTS

    def _get_statements(self, deployment_mode):
        return self.galaxy_statements[deployment_mode]

    def get_policy_statements(self, request, view):
        statements = self._get_statements(settings.GALAXY_DEPLOYMENT_MODE)
        return statements.get(self.NAME, [])

    def _get_rh_identity(self, request):
        if not isinstance(request.auth, dict):
            log.debug("No request rh_identity request.auth found for request %s", request)
            return False

        x_rh_identity = request.auth.get('rh_identity')
        if not x_rh_identity:
            return False

        return x_rh_identity

    # used by insights access policy
    def has_rh_entitlements(self, request, view, permission):

        x_rh_identity = self._get_rh_identity(request)

        if not x_rh_identity:
            log.debug("No x_rh_identity found when check entitlements for request %s for view %s",
                      request, view)
            return False

        entitlements = x_rh_identity.get('entitlements', {})
        entitlement = entitlements.get(settings.RH_ENTITLEMENT_REQUIRED, {})
        return entitlement.get('is_entitled', False)


class NamespaceAccessPolicy(UnauthenticatedCollectionAccessMixin, AccessPolicyBase):
    NAME = 'NamespaceViewSet'


class CollectionAccessPolicy(UnauthenticatedCollectionAccessMixin, AccessPolicyBase):
    NAME = 'CollectionViewSet'

    def can_update_collection(self, request, view, permission):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for update permissions
            return False
        collection = view.get_object()
        namespace = models.Namespace.objects.get(name=collection.namespace)
        return request.user.has_perm('galaxy.upload_to_namespace', namespace)

    def can_create_collection(self, request, view, permission):
        data = view._get_data(request)
        try:
            namespace = models.Namespace.objects.get(name=data['filename'].namespace)
        except models.Namespace.DoesNotExist:
            raise NotFound(_('Namespace in filename not found.'))
        return request.user.has_perm('galaxy.upload_to_namespace', namespace)

    def unauthenticated_collection_download_enabled(self, request, view, permission):
        return settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_DOWNLOAD


class CollectionRemoteAccessPolicy(AccessPolicyBase):
    NAME = 'CollectionRemoteViewSet'


class UserAccessPolicy(AccessPolicyBase):
    NAME = 'UserViewSet'

    def user_is_superuser(self, request, view, action):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for superuser
            return False
        user = view.get_object()
        return user.is_superuser

    def is_current_user(self, request, view, action):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for current user
            return False
        return request.user == view.get_object()


class MyUserAccessPolicy(UnauthenticatedCollectionAccessMixin, AccessPolicyBase):
    NAME = 'MyUserViewSet'

    def is_current_user(self, request, view, action):
        return request.user == view.get_object()


class SyncListAccessPolicy(AccessPolicyBase):
    NAME = 'SyncListViewSet'


class MySyncListAccessPolicy(AccessPolicyBase):
    NAME = 'MySyncListViewSet'

    def is_org_admin(self, request, view, permission):
        """Check the rhn_entitlement data to see if user is an org admin"""
        x_rh_identity = self._get_rh_identity(request)

        if not x_rh_identity:
            log.debug("No x_rh_identity found for request %s for view %s",
                      request, view)
            return False

        identity = x_rh_identity['identity']
        user = identity['user']
        return user.get('is_org_admin', False)


class TagsAccessPolicy(AccessPolicyBase):
    NAME = 'TagViewSet'


class TaskAccessPolicy(AccessPolicyBase):
    NAME = 'TaskViewSet'


class LoginAccessPolicy(AccessPolicyBase):
    NAME = 'LoginView'


class LogoutAccessPolicy(AccessPolicyBase):
    NAME = 'LogoutView'


class TokenAccessPolicy(AccessPolicyBase):
    NAME = 'TokenView'


class GroupAccessPolicy(AccessPolicyBase):
    NAME = 'GroupViewSet'


class DistributionAccessPolicy(AccessPolicyBase):
    NAME = 'DistributionViewSet'


class MyDistributionAccessPolicy(AccessPolicyBase):
    NAME = 'MyDistributionViewSet'


class ContainerRepositoryAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerRepositoryViewSet'


class ContainerReadmeAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerReadmeViewset'

    def has_container_namespace_perms(self, request, view, action, permission):
        readme = view.get_object()
        return (request.user.has_perm(permission)
                or request.user.has_perm(permission, readme.container.namespace))


class ContainerNamespaceAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerNamespaceViewset'


class ContainerRegistryRemoteAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerRegistryRemoteViewSet'


class ContainerRemoteAccessPolicy(AccessPolicyBase, NamespacedAccessPolicyMixin):
    NAME = 'ContainerRemoteViewSet'

    def has_distro_permission(self, request, view, action, permission):
        class FakeView:
            def __init__(self, obj):
                self.obj = obj

            def get_object(self):
                return self.obj

        # has_container_namespace_perms

        remote = view.get_object()
        repositories = remote.repository_set.all()

        # In theory there should never be more than one repository connected to a remote, but
        # the pulp apis don't prevent you from attaching as many remotes as you want to a repo.
        for repo in repositories:
            for distro in container_models.ContainerDistribution.objects.filter(repository=repo):
                dummy_view = FakeView(distro)
                if self.has_namespace_or_obj_perms(request, dummy_view, action, permission):
                    return True

        return False




from pulp_container.app.tasks.synchronize import synchronize as container_sync
from pulpcore.plugin.tasking import dispatch

from galaxy_ng.app import models


def launch_container_remote_sync(remote, registry, repository):
    for key, value in registry.get_connection_fields().items():
        setattr(remote, key, value)
        remote.save()

    return dispatch(
        container_sync,
        shared_resources=[remote],
        exclusive_resources=[repository],
        kwargs={
            "remote_pk": str(remote.pk),
            "repository_pk": str(repository.pk),
            "mirror": True,
        },
    )


def sync_all_repos_in_registry(registry_pk):
    registry = models.ContainerRegistryRemote.objects.get(pk=registry_pk)
    for remote_rel in models.ContainerRegistryRepos.objects.filter(registry=registry).all():
        remote = remote_rel.repository_remote

        for repo in remote.repository_set.all():
            repo.cast()
            launch_container_remote_sync(remote, registry, repo)



STANDALONE_STATEMENTS = {
    'NamespaceViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": ["list", "retrieve"],
            "principal": "anonymous",
            "effect": "allow",
            "condition": "unauthenticated_collection_access_enabled"
        },
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_or_obj_perms:galaxy.delete_namespace"
        },
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.add_namespace"
        },
        {
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_or_obj_perms:galaxy.change_namespace"
        },
    ],
    'CollectionViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": ["list", "retrieve"],
            "principal": "anonymous",
            "effect": "allow",
            "condition": "unauthenticated_collection_access_enabled"
        },
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:ansible.delete_collection",
        },
        {
            "action": ["download"],
            "principal": 'authenticated',
            "effect": "allow",
        },
        {
            "action": ["download"],
            "principal": 'anonymous',
            "effect": "allow",
            "condition": "unauthenticated_collection_download_enabled",
        },
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "can_create_collection"
        },
        {
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "can_update_collection"
        },
        {
            "action": "move_content",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:ansible.modify_ansible_repo_content"
        }
    ],
    'CollectionRemoteViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow"
        },
        {
            "action": ["sync", "update", "partial_update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:ansible.change_collectionremote"
        }
    ],
    'UserViewSet': [
        {
            "action": ["list"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.view_user"
        },
        {
            "action": ["retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.view_user"
        },
        {
            "action": "destroy",
            "principal": "*",
            "effect": "deny",
            "condition": ["user_is_superuser"]
        },
        {
            "action": "destroy",
            "principal": "*",
            "effect": "deny",
            "condition": ["is_current_user"]
        },
        {
            "action": "destroy",
            "principal": "*",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.delete_user"
        },
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.add_user"
        },
        {
            "action": ["update", "partial_update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.change_user"
        },
    ],
    'MyUserViewSet': [
        {
            "action": ["retrieve"],
            "principal": "anonymous",
            "effect": "allow",
            "condition": "unauthenticated_collection_access_enabled"
        },
        {
            "action": ["retrieve", "update", "partial_update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "is_current_user"
        },
    ],
    'SyncListViewSet': [
        #  disable synclists for on prem installations
        {
            "action": ["*"],
            "principal": "*",
            "effect": "deny",
        },
    ],
    'MySyncListViewSet': [
        #  disable synclists for on prem installations
        {
            "action": ["*"],
            "principal": "*",
            "effect": "deny",
        },
    ],
    'TaskViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    'TagViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    # LoginView, LogoutView and TokenView are all views instead of viewsets.
    # At the moment, DRF access policy doesn't seem to be able to correctly
    # determine the action for views, so we have to resort to using *
    'LoginView': [
        {
            "action": ['*'],
            "principal": "*",
            "effect": "allow",
        }
    ],
    'LogoutView': [
        {
            "action": ['*'],
            "principal": "*",
            "effect": "allow"
        }
    ],
    'TokenView': [
        {
            "action": ['*'],
            "principal": "authenticated",
            "effect": "allow"
        }
    ],
    'GroupViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.delete_group"
        },
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.add_group"
        },
        {
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.update_group"
        },
    ],
    'DistributionViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    'MyDistributionViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    # TODO: More specific permissions will be required here when pulp_container
    # RBAC is added
    'ContainerRepositoryViewSet': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:container.delete_containerrepository",
        },
    ],

    # The container readme can't just use the ContainerRepository access policies
    # because the readme viewset returns a readme object and not a container
    # repository object, which breaks has_model_or_obj_perms.
    'ContainerReadmeViewset': [
        {
            "action": ["retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": ["update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ("has_container_namespace_perms:"
                          "container.namespace_change_containerdistribution")
        },
    ],

    'ContainerNamespaceViewset': [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            "action": ["update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_or_obj_perms:container.change_containernamespace"
        },
    ],

    'ContainerRegistryRemoteViewSet': [
        # prevents deletion of registry
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.delete_containerregistryremote"
        },
        # allows authenticated users to VIEW registries
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        # allows users with model create permissions to create new registries
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.add_containerregistryremote"
        },
        {
            "action": "sync",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.change_containerregistryremote"
        },
        # allows users with model create permissions to update registries
        {
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.change_containerregistryremote"
        },
        {
            "action": "index_execution_environments",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:galaxy.change_containerregistryremote"
        },
    ],

    'ContainerRemoteViewSet': [
        # The permissions for creating namespaces are used as a proxy for creating containers,
        # so allow users to create container remotes if they have create namespace permissions.
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:container.add_containernamespace"
        },
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
        {
            # Permissions for containers are controlled via container distributions and namespaces.
            # Since the remote controls what will go into the container distribution, reuse the
            # same permissions here.
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_distro_permission:container.change_containerdistribution"
        },
        {
            "action": "sync",
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_distro_permission:container.change_containerdistribution"
        },
    ],
}



"""Utilities for tests for the galaxy plugin."""
import os
from functools import partial
import random
import requests
from unittest import SkipTest
from tempfile import NamedTemporaryFile

from pulp_smash import api, config, selectors
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, PulpTestCase
from pulp_smash.pulp3.utils import (
    gen_remote,
    gen_repo,
    get_content,
    require_pulp_3,
    require_pulp_plugins,
    sync,
)

from galaxy_ng.tests.functional.constants import (
    GALAXY_CONTENT_NAME,
    GALAXY_CONTENT_PATH,
    GALAXY_FIXTURE_URL,
    GALAXY_PUBLICATION_PATH,
    GALAXY_REMOTE_PATH,
    GALAXY_REPO_PATH,
    GALAXY_URL,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ArtifactsApi,
    TasksApi,
)
from pulpcore.client.galaxy_ng import (
    ApiClient as GalaxyApiClient,
    ApiContentV3CollectionsApi,
    ApiContentV3SyncApi,
    ApiContentV3SyncConfigApi,
    ApiV3NamespacesApi,
    ApiUiV1ExecutionEnvironmentsRepositoriesContentTagsApi as ContainerRepositoryEndpointApi,
    ApiUiV1ExecutionEnvironmentsRepositoriesApi as ContainerRepositoryApi,
    ApiUiV1ExecutionEnvironmentsRemotesApi,
    ApiUiV1ExecutionEnvironmentsRegistriesApi,
    ApiUiV1ExecutionEnvironmentsRepositoriesContentSyncApi,
    ApiUiV1ExecutionEnvironmentsRegistriesSyncApi,
    ApiUiV1ExecutionEnvironmentsRepositoriesContentImagesApi as ContainerImagesAPI,
)


configuration = config.get_config().get_bindings_config()


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if galaxy_ng isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({"galaxy"}, SkipTest)


def gen_galaxy_client():
    """Return an OBJECT for galaxy client."""
    return GalaxyApiClient(configuration)


def gen_galaxy_remote(url=GALAXY_FIXTURE_URL, **kwargs):
    """Return a semi-random dict for use in creating a galaxy Remote.

    :param url: The URL of an external content source.
    """
    # FIXME: Add any fields specific to a galaxy remote here
    return gen_remote(url, **kwargs)


def get_galaxy_content_paths(repo, version_href=None):
    """Return the relative path of content units present in a galaxy repository.

    :param repo: A dict of information about the repository.
    :param version_href: The repository version to read.
    :returns: A dict of lists with the paths of units present in a given repository.
        Paths are given as pairs with the remote and the local version for different content types.
    """
    # FIXME: The "relative_path" is actually a file path and name
    # It's just an example -- this needs to be replaced with an implementation that works
    # for repositories of this content type.
    return {
        GALAXY_CONTENT_NAME: [
            (content_unit["relative_path"], content_unit["relative_path"])
            for content_unit in get_content(repo, version_href)[GALAXY_CONTENT_NAME]
        ],
    }


def gen_galaxy_content_attrs(artifact):
    """Generate a dict with content unit attributes.

    :param artifact: A dict of info about the artifact.
    :returns: A semi-random dict for use in creating a content unit.
    """
    # FIXME: Add content specific metadata here.
    return {"_artifact": artifact["pulp_href"]}


def populate_pulp(cfg, url=GALAXY_FIXTURE_URL):
    """Add galaxy contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp application.
    :param url: The galaxy repository URL. Defaults to
        :data:`pulp_smash.constants.GALAXY_FIXTURE_URL`
    :returns: A list of dicts, where each dict describes one galaxy content in Pulp.
    """
    client = api.Client(cfg, api.json_handler)
    remote = {}
    repo = {}
    try:
        remote.update(client.post(GALAXY_REMOTE_PATH, gen_galaxy_remote(url)))
        repo.update(client.post(GALAXY_REPO_PATH, gen_repo()))
        sync(cfg, remote, repo)
    finally:
        if remote:
            client.delete(remote["pulp_href"])
        if repo:
            client.delete(repo["pulp_href"])
    return client.get(GALAXY_CONTENT_PATH)["results"]


def publish(cfg, repo, version_href=None):
    """Publish a repository.
    :param pulp_smash.config.PulpSmashConfig cfg: Information about the Pulp
        host.
    :param repo: A dict of information about the repository.
    :param version_href: A href for the repo version to be published.
    :returns: A publication. A dict of information about the just created
        publication.
    """
    if version_href:
        body = {"repository_version": version_href}
    else:
        body = {"repository": repo["pulp_href"]}

    client = api.Client(cfg, api.json_handler)
    call_report = client.post(GALAXY_PUBLICATION_PATH, body)
    tasks = tuple(api.poll_spawned_tasks(cfg, call_report))
    return client.get(tasks[-1]["created_resources"][0])


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""

core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)


def gen_artifact(url=GALAXY_URL):
    """Creates an artifact."""
    response = requests.get(url)
    with NamedTemporaryFile() as temp_file:
        temp_file.write(response.content)
        temp_file.flush()
        artifact = ArtifactsApi(core_client).create(file=temp_file.name)
        return artifact.to_dict()


class TestCaseUsingBindings(PulpTestCase):
    """A parent TestCase that instantiates the various bindings used throughout tests."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_galaxy_client()
        cls.smash_client = api.Client(cls.cfg, api.smart_handler)
        cls.namespace_api = ApiV3NamespacesApi(cls.client)
        cls.collections_api = ApiContentV3CollectionsApi(cls.client)
        cls.sync_config_api = ApiContentV3SyncConfigApi(cls.client)
        cls.sync_api = ApiContentV3SyncApi(cls.client)
        cls.container_repo_tags_api = ContainerRepositoryEndpointApi(cls.client)
        cls.container_repo_api = ContainerRepositoryApi(cls.client)
        cls.container_remotes_api = ApiUiV1ExecutionEnvironmentsRemotesApi(cls.client)
        cls.container_registries_api = ApiUiV1ExecutionEnvironmentsRegistriesApi(cls.client)
        cls.container_remote_sync_api = ApiUiV1ExecutionEnvironmentsRepositoriesContentSyncApi(cls.client)
        cls.container_registry_sync_api = ApiUiV1ExecutionEnvironmentsRegistriesSyncApi(cls.client)
        cls.container_images_api = ContainerImagesAPI(cls.client)
        cls.get_ansible_cfg_before_test()
        cls.galaxy_api_prefix = os.getenv("PULP_GALAXY_API_PATH_PREFIX", "/api/galaxy").rstrip("/")

    def tearDown(self):
        """Clean class-wide variable."""
        with open("ansible.cfg", "w") as f:
            f.write(self.previous_ansible_cfg)
        delete_orphans()

    @classmethod
    def get_token(cls):
        """Get a Galaxy NG token."""
        return cls.smash_client.post(f"{cls.galaxy_api_prefix}/v3/auth/token/")["token"]

    @classmethod
    def get_ansible_cfg_before_test(cls):
        """Update ansible.cfg to use the given base_path."""
        with open("ansible.cfg", "r") as f:
            cls.previous_ansible_cfg = f.read()

    def update_ansible_cfg(self, base_path):
        """Update ansible.cfg to use the given base_path."""
        token = self.get_token()
        ansible_cfg = (
            f"{self.previous_ansible_cfg}\n"
            "[galaxy]\n"
            "server_list = community_repo\n"
            "\n"
            "[galaxy_server.community_repo]\n"
            f"url={self.cfg.get_base_url()}"
            f"{self.galaxy_api_prefix}/content/{base_path}/\n"
            f"token={token}"
        )
        with open("ansible.cfg", "w") as f:
            f.write(ansible_cfg)

    def sync_repo(self, requirements_file, **kwargs):
        """Sync a repository with a given requirements_file"""
        repo_name = kwargs.get("repo_name", "community")
        url = kwargs.get("url", "https://galaxy.ansible.com/api/")

        self.sync_config_api.update(
            repo_name,
            {
                "url": f"{url}",
                "requirements_file": f"{requirements_file}",
            },
        )

        response = self.sync_api.sync(repo_name)
        monitor_task(f"/pulp/api/v3/tasks/{response.task}/")


    def delete_namespace(self, namespace_name):
        """Delete a Namespace"""
        # namespace_api does not support delete, so we can use the smash_client directly
        self.smash_client.delete(
            f"{self.galaxy_api_prefix}/v3/namespaces/{namespace_name}"
        )

    def delete_collection(self, collection_namespace, collection_name):
        """Delete a Collection"""
        self.smash_client.delete(
            f"{self.galaxy_api_prefix}/v3/collections/{collection_namespace}/{collection_name}/"
        )


class ContainerTestCaseUsingBindings(TestCaseUsingBindings):
    @classmethod
    def setUpClass(cls):
        if os.getenv("TEST") == "azure":
            raise SkipTest("Container Registry disabled")
        super().setUpClass()